"""The interface for the database."""
from __future__ import annotations

import abc
import enum
import pathlib
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import Any, Callable, Iterable, Iterator, Literal, Optional, Sequence, Union

import pandas as pd
from pydantic import (
  BaseModel,
  ConfigDict,
  SerializeAsAny,
  StrictBool,
  StrictBytes,
  StrictFloat,
  StrictInt,
  StrictStr,
  field_validator,
)
from pydantic import Field as PydanticField
from typing_extensions import TypeAlias

from ..auth import UserInfo
from ..config import (
  DatasetConfig,
  DatasetSettings,
  DatasetUISettings,
  EmbeddingConfig,
  SignalConfig,
  get_dataset_config,
)
from ..dataset_format import DatasetFormat, DatasetFormatInputSelector
from ..env import get_project_dir
from ..project import read_project_config, update_project_dataset_settings
from ..schema import (
  PATH_WILDCARD,
  ROWID,
  STRING,
  STRING_SPAN,
  Bin,
  EmbeddingInputType,
  Field,
  Item,
  MapFn,
  Path,
  PathKey,
  PathTuple,
  Schema,
  SpanVector,
  change_const_to_enum,
  normalize_path,
)
from ..signal import (
  Signal,
  TextEmbeddingSignal,
  TopicFn,
  resolve_signal,
)
from ..signals.concept_scorer import ConceptSignal
from ..source import Source, resolve_source
from ..tasks import TaskExecutionType, TaskId

# Threshold for rejecting certain queries (e.g. group by) for columns with large cardinality.
TOO_MANY_DISTINCT = 1_000_000
SAMPLE_AVG_TEXT_LENGTH = 1000
MAX_TEXT_LEN_DISTINCT_COUNT = 250
# Threshold for promoting a column to a media field.
MEDIA_AVG_TEXT_LEN = 150

# Special name for the deleted label.
DELETED_LABEL_NAME = '__deleted__'


class SelectRowsResult:
  """The result of a select rows query."""

  def __init__(self, df: pd.DataFrame, total_num_rows: int) -> None:
    """Initialize the result."""
    self._df = df
    self.total_num_rows = total_num_rows
    self._next_iter: Optional[Iterator] = None

  def __iter__(self) -> Iterator:
    # Replace NaT timestamps with Nones.
    df = self._df.replace({pd.NaT: None})
    return (row.to_dict() for _, row in df.iterrows())

  def __next__(self) -> Item:
    if not self._next_iter:
      self._next_iter = self.__iter__()
    try:
      return next(self._next_iter)
    except StopIteration:
      self._next_iter = None
      raise

  def df(self) -> pd.DataFrame:
    """Convert the result to a pandas DataFrame."""
    return self._df


class StatsResult(BaseModel):
  """The result of a stats() query."""

  path: PathTuple
  # The number of leaf values.
  total_count: int
  # The approximate number of distinct leaf values.
  approx_count_distinct: int

  # Defined for ordinal features.
  min_val: Optional[Union[float, datetime]] = None
  max_val: Optional[Union[float, datetime]] = None

  # Defined for text features.
  avg_text_length: Optional[float] = None


class MediaResult(BaseModel):
  """The result of a media() query."""

  data: bytes


BinaryOp = Literal['equals', 'not_equal', 'greater', 'greater_equal', 'less', 'less_equal']
StringOp = Literal['length_longer', 'length_shorter', 'ilike', 'regex_matches', 'not_regex_matches']
UnaryOp = Literal['exists', 'not_exists', None]
ListOp = Literal['in', None]
RawSqlOp = Literal['raw_sql']

BINARY_OPS = set(['equals', 'not_equal', 'greater', 'greater_equal', 'less', 'less_equal'])
STRING_OPS = set(['length_longer', 'length_shorter', 'ilike', 'regex_matches', 'not_regex_matches'])
UNARY_OPS = set(['exists', 'not_exists'])
LIST_OPS = set(['in'])
RAW_SQL_OPS = set(['raw_sql'])
ALL_OPS = sorted(BINARY_OPS | STRING_OPS | UNARY_OPS | LIST_OPS | RAW_SQL_OPS)


class SortOrder(str, enum.Enum):
  """The sort order for a database query."""

  DESC = 'DESC'
  ASC = 'ASC'


class GroupsSortBy(str, enum.Enum):
  """The sort for groups queries.

  Either "count" which sorts by the count of feature value, or "value" which sorts by the
  feature value itself.
  """

  COUNT = 'count'
  VALUE = 'value'


class SortResult(BaseModel):
  """The information about what is sorted after combining searches and explicit sorts."""

  # The column that was sorted.
  path: PathTuple
  # The sort order.
  order: SortOrder
  # The alias of the column if it was aliased.
  alias: Optional[str] = None
  # The search index if the sort is by a search.
  search_index: Optional[int] = None


class SearchResultInfo(BaseModel):
  """The resulting sort order returned by the select rows schema."""

  # The input path to the search.
  search_path: PathTuple
  # The resulting column that was searched.
  result_path: PathTuple
  # The alias of the UDF.
  alias: Optional[str] = None


class SelectRowsSchemaUDF(BaseModel):
  """The UDF for a select rows schema query."""

  path: PathTuple
  alias: Optional[str] = None


class SelectRowsSchemaResult(BaseModel):
  """The result of a select rows schema query."""

  data_schema: Schema
  udfs: list[SelectRowsSchemaUDF] = []
  search_results: list[SearchResultInfo] = []
  sorts: Optional[list[SortResult]] = None


class Column(BaseModel):
  """A column in the dataset."""

  path: PathTuple
  alias: Optional[str] = None  # This is the renamed column during querying and response.

  # Defined when the feature is another column.
  signal_udf: Optional[Signal] = None

  def __init__(
    self,
    path: Path,
    alias: Optional[str] = None,
    signal_udf: Optional[Signal] = None,
    **kwargs: Any,
  ):
    """Initialize a column. We override __init__ to allow positional arguments for brevity."""
    super().__init__(path=normalize_path(path), alias=alias, signal_udf=signal_udf, **kwargs)

  @field_validator('signal_udf', mode='before')
  @classmethod
  def parse_signal_udf(cls, signal_udf: Optional[dict]) -> Optional[Signal]:
    """Parse a signal to its specific subclass instance."""
    if not signal_udf:
      return None
    return resolve_signal(signal_udf)


ColumnId = Union[Path, Column]


class DatasetManifest(BaseModel):
  """The manifest for a dataset."""

  namespace: str
  dataset_name: str
  data_schema: Schema
  source: SerializeAsAny[Source]
  # Number of items in the dataset.
  num_items: int

  # The format of the dataset, automatically inferred.
  dataset_format: Optional[DatasetFormat] = None

  @field_validator('source', mode='before')
  @classmethod
  def parse_source(cls, source: dict) -> Source:
    """Parse a source to its specific subclass instance."""
    return resolve_source(source)


def column_from_identifier(column: ColumnId) -> Column:
  """Create a column from a column identifier."""
  if isinstance(column, Column):
    return column.model_copy()
  return Column(path=column)


FeatureValue = Union[StrictBool, StrictInt, StrictFloat, StrictStr, StrictBytes, datetime]
FeatureListValue = list[StrictStr]
BinaryFilterTuple = tuple[Path, BinaryOp, FeatureValue]
StringFilterTuple = tuple[Path, StringOp, FeatureValue]
ListFilterTuple = tuple[Path, ListOp, FeatureListValue]
UnaryFilterTuple = tuple[Path, UnaryOp]

FilterOp = Union[BinaryOp, StringOp, UnaryOp, ListOp, RawSqlOp]


class SelectGroupsResult(BaseModel):
  """The result of a select groups query."""

  too_many_distinct: bool
  counts: list[tuple[Optional[FeatureValue], int]]
  bins: Optional[list[Bin]] = None


class PivotResultOuterGroup(BaseModel):
  """The outer group of a pivot result."""

  value: Optional[str]
  count: int
  inner: list[tuple[Optional[str], int]]


class PivotResult(BaseModel):
  """The result of a pivot query."""

  outer_groups: list[PivotResultOuterGroup]
  too_many_distinct: bool = False


class Filter(BaseModel):
  """A filter on a column."""

  path: PathTuple
  op: FilterOp = PydanticField(description=f'The type of filter. Available filters are {ALL_OPS}')
  value: Optional[Union[FeatureValue, FeatureListValue]] = None


FilterLike: TypeAlias = Union[
  Filter, BinaryFilterTuple, StringFilterTuple, UnaryFilterTuple, ListFilterTuple
]

SearchValue = StrictStr


class KeywordSearch(BaseModel):
  """A keyword search query on a column."""

  path: Path
  query: SearchValue
  type: Literal['keyword'] = 'keyword'

  model_config = ConfigDict(json_schema_extra=change_const_to_enum('type', 'keyword'))


class SemanticSearch(BaseModel):
  """A semantic search on a column."""

  path: Path
  query: SearchValue
  embedding: str
  type: Literal['semantic'] = 'semantic'
  query_type: EmbeddingInputType = PydanticField(
    title='Query Type',
    default='document',
    description='The input type of the query, used for the query embedding.',
  )

  model_config = ConfigDict(json_schema_extra=change_const_to_enum('type', 'semantic'))


class ConceptSearch(BaseModel):
  """A concept search query on a column."""

  path: Path
  concept_namespace: str
  concept_name: str
  embedding: str
  type: Literal['concept'] = 'concept'

  model_config = ConfigDict(json_schema_extra=change_const_to_enum('type', 'concept'))


class MetadataSearch(BaseModel):
  """A metadata search query on a column."""

  path: Path
  op: FilterOp
  value: Optional[Union[FeatureValue, FeatureListValue]] = None
  type: Literal['metadata'] = 'metadata'

  model_config = ConfigDict(json_schema_extra=change_const_to_enum('type', 'metadata'))


Search = Union[ConceptSearch, SemanticSearch, KeywordSearch, MetadataSearch]


class Dataset(abc.ABC):
  """The database implementation to query a dataset."""

  namespace: str
  dataset_name: str
  project_dir: Union[str, pathlib.Path]

  def __init__(
    self, namespace: str, dataset_name: str, project_dir: Optional[Union[str, pathlib.Path]] = None
  ):
    """Initialize a dataset.

    Args:
      namespace: The dataset namespace.
      dataset_name: The dataset name.
      project_dir: The path to the project directory.
    """
    self.namespace = namespace
    self.dataset_name = dataset_name
    self.project_dir = project_dir or get_project_dir()

  @abc.abstractmethod
  def delete(self) -> None:
    """Deletes the dataset."""
    pass

  @abc.abstractmethod
  def manifest(self) -> DatasetManifest:
    """Return the manifest for the dataset."""
    pass

  def config(self) -> DatasetConfig:
    """Return the dataset config for this dataset."""
    manifest = self.manifest()
    project_config = read_project_config(get_project_dir())
    dataset_config = get_dataset_config(project_config, self.namespace, self.dataset_name)
    if not dataset_config:
      raise ValueError(
        f'Dataset "{self.namespace}/{self.dataset_name}" not found in project config.'
      )

    # Filter out invalid media paths.
    if dataset_config.settings and dataset_config.settings.ui.media_paths:
      dataset_config.settings.ui.media_paths = [
        p for p in dataset_config.settings.ui.media_paths if manifest.data_schema.has_field(p)
      ]

    return dataset_config

  def settings(self) -> DatasetSettings:
    """Return the persistent settings for the dataset."""
    settings = self.config().settings
    return settings or DatasetSettings()

  def update_settings(self, settings: DatasetSettings) -> None:
    """Update the persistent settings for the dataset."""
    update_project_dataset_settings(self.namespace, self.dataset_name, settings, self.project_dir)

  def add_media_field(self, path: Path, markdown: bool = False) -> None:
    """Add a media field to the dataset."""
    settings = self.settings()
    path = normalize_path(path)
    if not self.manifest().data_schema.has_field(path):
      raise ValueError(f'Field "{path}" does not exist in dataset.')
    field = self.manifest().data_schema.get_field(path)
    if field.dtype not in (STRING, STRING_SPAN):
      raise ValueError(f'Field "{path}" is not a string field.')
    if path not in settings.ui.media_paths:
      settings.ui.media_paths.append(path)
    if markdown:
      if path not in settings.ui.markdown_paths:
        settings.ui.markdown_paths.append(path)
    self.update_settings(settings)

  def remove_media_field(self, path: Path) -> None:
    """Remove a media field from the dataset."""
    settings = self.settings()
    path = normalize_path(path)
    if path in settings.ui.media_paths:
      settings.ui.media_paths.remove(path)
    if path in settings.ui.markdown_paths:
      settings.ui.markdown_paths.remove(path)
    self.update_settings(settings)

  @abc.abstractmethod
  def compute_signal(
    self,
    signal: Signal,
    path: Path,
    filters: Optional[Sequence[FilterLike]] = None,
    limit: Optional[int] = None,
    include_deleted: bool = False,
    overwrite: bool = False,
    task_id: Optional[TaskId] = None,
    remote: bool = False,
  ) -> None:
    """Compute a signal for a column.

    Args:
      signal: The signal to compute over the given columns.
      path: The leaf path to compute the signal on.
      filters: Filters to apply to the row; only matching rows will have the signal computed.
      limit: Limit the number of rows to compute the signal on.
      include_deleted: Whether to include deleted rows in the computation.
      overwrite: Whether to overwrite an existing signal computed at this path.
      task_id: The TaskManager `task_id` for this process run. This is used to update
        the progress of the task.
      remote: Whether to run the computation remotely on Lilac Garden.
    """
    pass

  @abc.abstractmethod
  def cluster(
    self,
    input: Union[Path, Callable[[Item], str], DatasetFormatInputSelector],
    output_path: Optional[Path] = None,
    min_cluster_size: int = 5,
    topic_fn: Optional[TopicFn] = None,
    overwrite: bool = False,
    remote: bool = False,
    task_id: Optional[TaskId] = None,
  ) -> None:
    """Compute clusters for a field of the dataset.

    Args:
      input: The path to the text field to cluster, a function that returns a string for each row
        in the dataset, or a `DatasetFormatInputSelector` that is a format-specific input selector,
        like `ShareGPT.human`.
      output_path: The name of the output path to write to. Defaults to the input path + ".cluster".
      min_cluster_size: The minimum number of docs in a cluster.
      topic_fn: A function that returns a topic summary for each cluster. It takes a list of
        (doc, membership_score) tuples and returns a single topic. This is used to compute the topic
        for a given cluster of docs. It defaults to a function that summarizes user's requests.
      overwrite: Whether to overwrite an existing output.
      remote: Whether to run the clustering remotely on Lilac Garden.
      task_id: The TaskManager `task_id` for this process run. This is used to update the progress
        of the task.

    """
    pass

  @abc.abstractmethod
  def compute_embedding(
    self,
    embedding: str,
    path: Path,
    filters: Optional[Sequence[FilterLike]] = None,
    limit: Optional[int] = None,
    include_deleted: bool = False,
    overwrite: bool = False,
    task_id: Optional[TaskId] = None,
    remote: bool = False,
  ) -> None:
    """Compute an embedding for a given field path."""
    pass

  def compute_concept(
    self,
    namespace: str,
    concept_name: str,
    embedding: str,
    path: Path,
    filters: Optional[Sequence[FilterLike]] = None,
    limit: Optional[int] = None,
    include_deleted: bool = False,
    overwrite: bool = False,
    task_id: Optional[TaskId] = None,
  ) -> None:
    """Compute concept scores for a given field path."""
    signal = ConceptSignal(namespace=namespace, concept_name=concept_name, embedding=embedding)
    self.compute_signal(
      signal,
      path,
      filters,
      limit,
      include_deleted,
      overwrite=overwrite,
      task_id=task_id,
    )

  @abc.abstractmethod
  def delete_signal(self, signal_path: Path) -> None:
    """Delete a computed signal from the dataset.

    Args:
      signal_path: The path holding the computed data of the signal.
    """
    pass

  @abc.abstractmethod
  def delete_column(self, path: Path) -> None:
    """Delete a computed column from the dataset.

    Args:
      path: The path of the computed column.
    """
    pass

  @abc.abstractmethod
  def select_groups(
    self,
    leaf_path: Path,
    filters: Optional[Sequence[FilterLike]] = None,
    sort_by: Optional[GroupsSortBy] = None,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
    limit: Optional[int] = None,
    bins: Optional[Union[Sequence[Bin], Sequence[float]]] = None,
  ) -> SelectGroupsResult:
    """Select grouped columns to power a histogram.

    Args:
      leaf_path: The leaf path to group by. The path can be a dot-seperated string path, or a tuple
                 of fields.
      filters: The filters to apply to the query.
      sort_by: What to sort by, either "count" or "value".
      sort_order: The sort order.
      limit: The maximum number of rows to return.
      bins: The bins to use when bucketizing a float column.

    Returns:
      A `SelectGroupsResult` iterator where each row is a group.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def pivot(
    self,
    outer_path: Path,
    inner_path: Path,
    filters: Optional[Sequence[FilterLike]] = None,
    sort_by: Optional[GroupsSortBy] = GroupsSortBy.COUNT,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
  ) -> PivotResult:
    """Generate a pivot table with counts over two fields (outer and inner).

    For each value of `outer_path` in the dataset, the pivot table will contain the count of
    that value, as well as the internal counts of all the values of `inner_path` for the given
    `outer_path` value.

    For example, if `outer_path` is `split=train|test` and `inner_path` is `source=wiki|cc` the
    pivot table will look like:

    ```
    [
      {
        value: 'train', count: 100, inner: [{value: 'wiki', count: 94}, {value: 'cc', count: 6}],
      },
      {
        value: 'test', count: 10, inner: [{value: 'wiki', count: 3}, {value: 'cc', count: 7}],
      },
    ]
    ```
    """
    raise NotImplementedError

  @abc.abstractmethod
  def select_rows(
    self,
    columns: Optional[Sequence[ColumnId]] = None,
    searches: Optional[Sequence[Search]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    sort_by: Optional[Sequence[Path]] = None,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
    limit: Optional[int] = 100,
    offset: Optional[int] = 0,
    resolve_span: bool = False,
    combine_columns: bool = False,
    include_deleted: bool = False,
    user: Optional[UserInfo] = None,
  ) -> SelectRowsResult:
    """Select a set of rows that match the provided filters, analogous to SQL SELECT.

    Args:
      columns: The columns to select. A column is an instance of `Column` which can either
        define a path to a feature, or a column with an applied Transform, e.g. a Concept. If none,
        it selects all columns.
      searches: The searches to apply to the query.
      filters: The filters to apply to the query.
      sort_by: An ordered list of what to sort by. When defined, this is a list of aliases of column
        names defined by the "alias" field in Column. If no alias is provided for a column, an
        automatic alias is generated by combining each path element with a "."
        For example: e.g. ('person', 'name') => person.name. For columns that are transform columns,
        an alias must be provided explicitly. When sorting by a (nested) list of values, the sort
        takes the minumum value when `sort_order` is `ASC`, and the maximum value when `sort_order`
        is `DESC`.
      sort_order: The sort order.
      limit: The maximum number of rows to return.
      offset: The offset to start returning rows from.
      resolve_span: Whether to resolve the span of the row.
      combine_columns: Whether to combine columns into a single object. The object will be pruned
        to only include sub-fields that correspond to the requested columns.
      include_deleted: Whether to include deleted rows in the query.
      user: The authenticated user, if auth is enabled and the user is logged in. This is used to
        apply ACL to the query, especially for concepts.

    Returns:
      A `SelectRowsResult` iterator with rows of `Item`s.
    """
    pass

  @abc.abstractmethod
  def select_rows_schema(
    self,
    columns: Optional[Sequence[ColumnId]] = None,
    sort_by: Optional[Sequence[Path]] = None,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
    searches: Optional[Sequence[Search]] = None,
    combine_columns: bool = False,
  ) -> SelectRowsSchemaResult:
    """Returns the schema of the result of `select_rows` above with the same arguments."""
    pass

  @abc.abstractmethod
  def add_labels(
    self,
    name: str,
    row_ids: Optional[Sequence[str]] = None,
    searches: Optional[Sequence[Search]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    sort_by: Optional[Sequence[Path]] = None,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
    limit: Optional[int] = None,
    offset: Optional[int] = 0,
    include_deleted: bool = False,
    value: Optional[str] = 'true',
  ) -> int:
    """Adds a label to a row, or a set of rows defined by searches and filters.

    Returns the number of added labels.
    """
    pass

  @abc.abstractmethod
  def get_embeddings(
    self, embedding: str, rowid: str, path: Union[PathKey, str]
  ) -> list[SpanVector]:
    """Returns the span-level embeddings associated with a specific row value.

    Args:
      embedding: The embedding name (e.g. `gte-small`, or `jina-v2-small`).
      rowid: The row id to get embeddings for.
      path: The path within a row to get embeddings for.
        - If the row is a struct, e.g. {person: {document: {text: ...}}} and we want the embeddings
        for text, the row path would be `person.document.text`.
        - If the row has a list of strings, e.g. {docs: [{text: ...}, {text: ...}]}, then the
          row path would be `docs.0.text` for the 1st doc and `docs.1.text` for the 2nd doc.

    Returns:
      A list of `ll.SpanVector` dicts holding the span coordinates and the embedding.
    """
    pass

  @abc.abstractmethod
  def get_label_names(self) -> list[str]:
    """Returns the list of label names that have been added to the dataset."""
    pass

  @abc.abstractmethod
  def remove_labels(
    self,
    name: str,
    row_ids: Optional[Sequence[str]] = None,
    searches: Optional[Sequence[Search]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    sort_by: Optional[Sequence[Path]] = None,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
    limit: Optional[int] = None,
    offset: Optional[int] = 0,
    include_deleted: bool = False,
  ) -> int:
    """Removes labels from a row, or a set of rows defined by searches and filters.

    Returns the number of removed labels.
    """
    pass

  def delete_rows(
    self,
    row_ids: Optional[Sequence[str]] = None,
    searches: Optional[Sequence[Search]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    sort_by: Optional[Sequence[Path]] = None,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
    limit: Optional[int] = None,
    offset: Optional[int] = 0,
  ) -> int:
    """Deletes rows from the dataset.

    Returns the number of deleted rows.
    """
    return self.add_labels(
      name=DELETED_LABEL_NAME,
      row_ids=row_ids,
      searches=searches,
      filters=filters,
      sort_by=sort_by,
      sort_order=sort_order,
      limit=limit,
      offset=offset,
    )

  def restore_rows(
    self,
    row_ids: Optional[Sequence[str]] = None,
    searches: Optional[Sequence[Search]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    sort_by: Optional[Sequence[Path]] = None,
    sort_order: Optional[SortOrder] = SortOrder.DESC,
    limit: Optional[int] = None,
    offset: Optional[int] = 0,
  ) -> int:
    """Undeletes rows from the dataset.

    Returns the number of restored rows.
    """
    return self.remove_labels(
      DELETED_LABEL_NAME,
      row_ids=row_ids,
      searches=searches,
      filters=filters,
      sort_by=sort_by,
      sort_order=sort_order,
      limit=limit,
      offset=offset,
      include_deleted=True,
    )

  @abc.abstractmethod
  def stats(self, leaf_path: Path, include_deleted: bool = False) -> StatsResult:
    """Compute stats for a leaf path.

    Args:
      leaf_path: The leaf path to compute stats for.
      include_deleted: Whether to include deleted rows in the stats.

    Returns:
      A StatsResult.
    """
    pass

  @abc.abstractmethod
  def media(self, item_id: str, leaf_path: Path) -> MediaResult:
    """Return the media for a leaf path.

    Args:
      item_id: The item id to get media for.
      leaf_path: The leaf path for the media.

    Returns:
      A MediaResult.
    """
    pass

  @abc.abstractmethod
  def map(
    self,
    map_fn: MapFn,
    input_path: Optional[Path] = None,
    output_path: Optional[Path] = None,
    overwrite: bool = False,
    resolve_span: bool = False,
    batch_size: Optional[int] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    limit: Optional[int] = None,
    sort_by: Optional[Path] = None,
    sort_order: Optional[SortOrder] = SortOrder.ASC,
    include_deleted: bool = False,
    num_jobs: int = 1,
    execution_type: TaskExecutionType = 'threads',
    embedding: Optional[str] = None,
    schema: Optional[Field] = None,
  ) -> Iterable[Item]:
    """Maps a function over all rows in the dataset and writes the result to a new column.

    Args:
      map_fn: A callable that takes a full row item dictionary, and returns an Item for the
        result. The result Item can be a primitive, like a string.
      input_path: The path to the input column to map over. If not specified, the map function will
        be called with the full row item dictionary. If specified, the map function will be called
        with the value at the given path, flattened. The output column will be written in the same
        shape as the input column, paralleling its nestedness.
      output_path: The name of the output path to write to. It is often useful to emit results
        next to the input, so they will get hierarchically shown in the UI.
      overwrite: Set to true to overwrite this column if it already exists. If this bit is False,
        an error will be thrown if the column already exists.
      resolve_span: Whether to resolve the spans into text before calling the map function.
      batch_size: If provided, the map function will be called with a list of rows. Useful for
        batching API requests or other expensive operations. If unspecified, the map will receive
        one item at a time.
      filters: Filters limiting the set of rows to map over. At the moment, we do not support
        incremental computations; the output column will be null for rows that do not match the
        filter, and there is no way to fill in those nulls without recomputing the entire map with
        a less restrictive filter and overwrite=True.
      limit: How many rows to map over. If not specified, all rows will be mapped over.
      sort_by: The path to sort by. If specified, the map will be called with rows sorted by this
        path. This is useful for map functions that need to maintain state across rows.
      sort_order: The sort order. Defaults to ascending.
      include_deleted: Whether to include deleted rows in the query.
      num_jobs: The number of jobs to shard the work, defaults to 1. When set to -1, the number of
        jobs will correspond to the number of processors. If `num_jobs` is greater than the number
        of processors, it split the work into `num_jobs` and distribute amongst processors.
      execution_type: The local execution type of the map. Either "threads" or "processes". Threads
        are better for network bound tasks like making requests to an external server, while
        processes are better for CPU bound tasks, like running a local LLM.
      embedding: If specified, the map function will be called with the `lilac.SpanVector`s
        of each item, instead of the original text. This is needed for embedding-based computations
        (e.g. clustering).
      schema: The schema for the output of the map function. If not provided, the schema will be
        auto inferred.

    Returns:
      An iterable of items that are the result of map. The result item does not have the column name
      as part of the dictionary, it is exactly what is returned from the map.
    """
    pass

  def transform(
    self,
    transform_fn: Callable[[Iterator[Item]], Iterator[Item]],
    input_path: Optional[Path] = None,
    output_path: Optional[Path] = None,
    overwrite: bool = False,
    resolve_span: bool = False,
    filters: Optional[Sequence[FilterLike]] = None,
    limit: Optional[int] = None,
    sort_by: Optional[Path] = None,
    sort_order: Optional[SortOrder] = SortOrder.ASC,
    embedding: Optional[str] = None,
    schema: Optional[Field] = None,
  ) -> Iterable[Item]:
    """Transforms the entire dataset (or a column) and writes the result to a new column.

    Args:
      transform_fn: A callable that takes an iterable of all items in the dataset, and returns an
      iterable of the same length for the result.
      input_path: The path to the input column to map over. If not specified, the map function will
        be called with the full row item dictionary. If specified, the map function will be called
        with the value at the given path, flattened. The output column will be written in the same
        shape as the input column, paralleling its nestedness.
      output_path: The name of the output path to write to. It is often useful to emit results
        next to the input, so they will get hierarchically shown in the UI.
      overwrite: Set to true to overwrite this column if it already exists. If this bit is False,
        an error will be thrown if the column already exists.
      resolve_span: Whether to resolve the spans into text before calling the map function.
      filters: Filters limiting the set of rows to map over. At the moment, we do not support
        incremental computations; the output column will be null for rows that do not match the
        filter, and there is no way to fill in those nulls without recomputing the entire map with
        a less restrictive filter and overwrite=True.
      limit: How many rows to map over. If not specified, all rows will be mapped over.
      sort_by: The path to sort by. If specified, the map will be called with rows sorted by this
        path. This is useful for map functions that need to maintain state across rows.
      sort_order: The sort order. Defaults to ascending.
      embedding: If specified, the transform function will be called with the `lilac.SpanVector`s
        of each item, instead of the original text. This is needed for embedding-based computations
        (e.g. clustering).
      schema: The schema for the output of the map function. If not provided, the schema will be
        auto inferred.
    """
    return self.map(
      map_fn=transform_fn,
      input_path=input_path,
      output_path=output_path,
      overwrite=overwrite,
      resolve_span=resolve_span,
      batch_size=-1,
      filters=filters,
      limit=limit,
      sort_by=sort_by,
      sort_order=sort_order,
      num_jobs=1,
      execution_type='threads',
      embedding=embedding,
      schema=schema,
    )

  @abc.abstractmethod
  def to_json(
    self,
    filepath: Union[str, pathlib.Path],
    jsonl: bool = True,
    columns: Optional[Sequence[ColumnId]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    include_labels: Optional[Sequence[str]] = None,
    exclude_labels: Optional[Sequence[str]] = None,
  ) -> None:
    """Export the dataset to a JSON file.

    Args:
      filepath: The path to the file to export to.
      jsonl: Whether to export to JSONL or JSON.
      columns: The columns to export.
      filters: The filters to apply to the query.
      include_labels: The labels to include in the export.
      exclude_labels: The labels to exclude in the export.
    """
    pass

  @abc.abstractmethod
  def to_pandas(
    self,
    columns: Optional[Sequence[ColumnId]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    include_labels: Optional[Sequence[str]] = None,
    exclude_labels: Optional[Sequence[str]] = None,
  ) -> pd.DataFrame:
    """Export the dataset to a pandas DataFrame.

    Args:
      columns: The columns to export.
      filters: The filters to apply to the query.
      include_labels: The labels to include in the export.
      exclude_labels: The labels to exclude in the export.
    """
    pass

  @abc.abstractmethod
  def to_parquet(
    self,
    filepath: Union[str, pathlib.Path],
    columns: Optional[Sequence[ColumnId]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    include_labels: Optional[Sequence[str]] = None,
    exclude_labels: Optional[Sequence[str]] = None,
  ) -> None:
    """Export the dataset to a parquet file.

    Args:
      filepath: The path to the file to export to.
      columns: The columns to export.
      filters: The filters to apply to the query.
      include_labels: The labels to include in the export.
      exclude_labels: The labels to exclude in the export.
    """
    pass

  @abc.abstractmethod
  def to_csv(
    self,
    filepath: Union[str, pathlib.Path],
    columns: Optional[Sequence[ColumnId]] = None,
    filters: Optional[Sequence[FilterLike]] = None,
    include_labels: Optional[Sequence[str]] = None,
    exclude_labels: Optional[Sequence[str]] = None,
  ) -> None:
    """Export the dataset to a csv file.

    Args:
      filepath: The path to the file to export to.
      columns: The columns to export.
      filters: The filters to apply to the query.
      include_labels: The labels to include in the export.
      exclude_labels: The labels to exclude in the export.
    """
    pass

  def petals(self) -> dict[PathTuple, Field]:
    """Return the leafs of the dataset."""
    return self.manifest().data_schema.leafs


def default_settings(dataset: Dataset) -> DatasetSettings:
  """Gets the default settings for a dataset."""
  schema = dataset.manifest().data_schema
  leaf_paths = [
    path for path, field in schema.leafs.items() if field.dtype == STRING and path != (ROWID,)
  ]
  pool = ThreadPoolExecutor()
  stats: list[StatsResult] = list(pool.map(lambda leaf: dataset.stats(leaf), leaf_paths))
  media_paths: list[PathTuple] = [
    s.path for s in stats if s.avg_text_length and s.avg_text_length >= MEDIA_AVG_TEXT_LEN
  ]
  if not media_paths:
    # If there are no long text fields, pick the longest of the short text field.
    sorted_stats = sorted(
      [s for s in stats if s.avg_text_length], key=lambda s: s.avg_text_length or -1.0, reverse=True
    )
    if sorted_stats:
      media_paths = [sorted_stats[0].path]

  return DatasetSettings(ui=DatasetUISettings(media_paths=media_paths))


def config_from_dataset(dataset: Dataset) -> DatasetConfig:
  """Computes a DatasetConfig from a manifest.

  NOTE: This is only used for backwards compatibility. Once we remove the back-compat logic, this
  method can be removed.
  """
  manifest = dataset.manifest()
  all_fields = manifest.data_schema.all_fields

  all_signals = [
    (path, resolve_signal(field.signal)) for (path, field) in all_fields if field.signal
  ]
  signals: list[tuple[PathTuple, Signal]] = []
  embeddings: list[tuple[PathTuple, TextEmbeddingSignal]] = []
  for path, s in all_signals:
    source_path = path[0:-1]  # Remove the signal name from the path.
    if isinstance(s, TextEmbeddingSignal):
      embeddings.append((source_path, s))
    else:
      signals.append((source_path, s))

  return DatasetConfig(
    namespace=manifest.namespace,
    name=manifest.dataset_name,
    source=manifest.source,
    signals=[SignalConfig(path=path, signal=signal) for path, signal in signals],
    embeddings=[
      EmbeddingConfig(path=path, embedding=embedding.name) for path, embedding in embeddings
    ],
    settings=default_settings(dataset),
  )


def make_signal_parquet_id(
  signal: Signal, source_path: PathTuple, is_computed_signal: Optional[bool] = False
) -> str:
  """Return a unique identifier for this parquet table."""
  # Remove the wildcards from the parquet id since they are implicit.
  path = [*[p for p in source_path if p != PATH_WILDCARD], signal.key(is_computed_signal)]
  return '.'.join(path)


def get_map_parquet_id(output_path: PathTuple) -> str:
  """Return a unique identifier for this parquet table."""
  # Remove the wildcards from the parquet id since they are implicit.
  return 'map.' + '.'.join(output_path)



"""Test our public REST API."""
import os
from time import sleep

from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from .auth import (
  AuthenticationInfo,
  ConceptUserAccess,
  DatasetUserAccess,
  UserAccess,
  UserInfo,
  get_session_user,
)
from .server import app, start_server, stop_server

client = TestClient(app)


def test_auth_admin(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ADMIN_EMAILS': 'test@test.com,test2@test2.com'})

  @mocker.patch('modal.config.Config')
  class Config:
    def to_dict(self) -> dict:
      return {}

  # Override the session user so we make them an admin.
  def admin_user() -> UserInfo:
    return UserInfo(
      id='1', email='test@test.com', name='test', given_name='test', family_name='test'
    )

  app.dependency_overrides[get_session_user] = admin_user

  url = '/auth_info'
  response = client.get(url)
  assert response.status_code == 200
  assert AuthenticationInfo.model_validate(response.json()) == AuthenticationInfo(
    user=admin_user(),
    access=UserAccess(
      is_admin=True,
      create_dataset=True,
      dataset=DatasetUserAccess(
        compute_signals=True,
        delete_dataset=True,
        delete_signals=True,
        update_settings=True,
        create_label_type=True,
        edit_labels=True,
        label_all=True,
        delete_rows=True,
        execute_remotely=False,
      ),
      concept=ConceptUserAccess(delete_any_concept=True),
    ),
    auth_enabled=True,
  )


def test_auth_nonadmin(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ADMIN_EMAILS': 'test@test.com,test2@test2.com'})

  # Override the session user so we make them an admin.
  def user() -> UserInfo:
    return UserInfo(
      id='1', email='test_user@test.com', name='test', given_name='test', family_name='test'
    )

  app.dependency_overrides[get_session_user] = user

  url = '/auth_info'
  response = client.get(url)
  assert response.status_code == 200
  assert AuthenticationInfo.model_validate(response.json()) == AuthenticationInfo(
    user=user(),
    access=UserAccess(
      is_admin=False,
      create_dataset=False,
      dataset=DatasetUserAccess(
        compute_signals=False,
        delete_dataset=False,
        delete_signals=False,
        update_settings=False,
        create_label_type=False,
        edit_labels=False,
        label_all=True,
        delete_rows=False,
        execute_remotely=False,
      ),
      concept=ConceptUserAccess(delete_any_concept=False),
    ),
    auth_enabled=True,
  )


def test_auth_nonadmin_edit_labels(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ADMIN_EMAILS': 'admin@test.com'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_USER_EDIT_LABELS': 'True'})

  # Override the session user so we make them an admin.
  def user() -> UserInfo:
    return UserInfo(
      id='1', email='test@test.com', name='test', given_name='test', family_name='test'
    )

  app.dependency_overrides[get_session_user] = user

  url = '/auth_info'
  response = client.get(url)
  assert response.status_code == 200
  assert AuthenticationInfo.model_validate(response.json()) == AuthenticationInfo(
    user=user(),
    access=UserAccess(
      is_admin=False,
      create_dataset=False,
      dataset=DatasetUserAccess(
        compute_signals=False,
        delete_dataset=False,
        delete_signals=False,
        update_settings=False,
        create_label_type=False,
        # Users can edit labels (add and remove) but not create label types.
        edit_labels=True,
        label_all=True,
        delete_rows=False,
        execute_remotely=False,
      ),
      concept=ConceptUserAccess(delete_any_concept=False),
    ),
    auth_enabled=True,
  )


def test_auth_nonadmin_label_all(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ADMIN_EMAILS': 'admin@test.com'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_USER_EDIT_LABELS': 'True'})
  mocker.patch.dict(os.environ, {'LILAC_AUTH_USER_DISABLE_LABEL_ALL': 'True'})

  # Override the session user so we make them an admin.
  def user() -> UserInfo:
    return UserInfo(
      id='1', email='test@test.com', name='test', given_name='test', family_name='test'
    )

  app.dependency_overrides[get_session_user] = user

  url = '/auth_info'
  response = client.get(url)
  assert response.status_code == 200
  assert AuthenticationInfo.model_validate(response.json()) == AuthenticationInfo(
    user=user(),
    access=UserAccess(
      is_admin=False,
      create_dataset=False,
      dataset=DatasetUserAccess(
        compute_signals=False,
        delete_dataset=False,
        delete_signals=False,
        update_settings=False,
        create_label_type=False,
        # Users can edit labels (add and remove) but not create label types.
        edit_labels=True,
        # Users cannot use the label all feature.
        label_all=False,
        delete_rows=False,
        execute_remotely=False,
      ),
      concept=ConceptUserAccess(delete_any_concept=False),
    ),
    auth_enabled=True,
  )


def test_start_and_stop_server() -> None:
  start_server()
  sleep(1)
  stop_server()



"""Utils for transformer embeddings."""

import os
from typing import TYPE_CHECKING, Optional, Union

from ..utils import log

if TYPE_CHECKING:
  from sentence_transformers import SentenceTransformer
  from transformers import AutoModel

# This is not the literal pytorch batch dimension, but rather the batch of sentences passed to the
# model.encode function. The model will split this batch into smaller batches after sorting by text
# length (for performance reasons). A larger batch size gives sentence_transformer more
# opportunities to minimize padding by grouping similar sentence lengths together.
SENTENCE_TRANSFORMER_BATCH_SIZE = 1024

# We're using joblib, which uses spawning, not forking. So it should be safe to hardcode this to
# true without deadlocks.
os.environ['TOKENIZERS_PARALLELISM'] = 'true'


def setup_model_device(
  model: Union['SentenceTransformer', 'AutoModel'],
  model_name: str,
) -> Union['SentenceTransformer', 'AutoModel']:
  """Prepare a transformer model to run on the most efficient device / backend."""
  try:
    import torch.backends.mps
  except ImportError:
    raise ImportError(
      'Could not import the "torch" python package. ' 'Please install it with `pip install "torch".'
    )
  preferred_device: Optional[str] = None
  if torch.backends.mps.is_available():
    preferred_device = 'mps'
  elif torch.cuda.is_available():
    preferred_device = 'cuda'
  elif not torch.backends.mps.is_built():
    log('MPS not available because the current PyTorch install was not built with MPS enabled.')

  if preferred_device:
    model = model.to(preferred_device)

  log(f'{model_name} using device: {model.device}')

  return model



"""Tests for dataset.compute_signal()."""

from datetime import datetime
from typing import Iterable

import pytest
from freezegun import freeze_time
from pandas import Timestamp
from pytest_mock import MockerFixture

from ..schema import PATH_WILDCARD, ROWID, Item, field, schema
from ..source import clear_source_registry, register_source
from .dataset import DELETED_LABEL_NAME, DatasetManifest, SelectGroupsResult, SortOrder
from .dataset_test_utils import TestDataMaker, TestSource

TEST_ITEMS: list[Item] = [{'str': 'a', 'int': 1}, {'str': 'b', 'int': 2}, {'str': 'c', 'int': 3}]

TEST_TIME = datetime(2023, 8, 15, 1, 23, 45)


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_source(TestSource)

  # Unit test runs.
  yield

  # Teardown.
  clear_source_registry()


@freeze_time(TEST_TIME)
def test_add_single_label(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  num_labels = dataset.add_labels('test_label', filters=[(ROWID, 'equals', '1')])
  assert num_labels == 1
  assert dataset.manifest() == DatasetManifest(
    source=TestSource(),
    namespace='test_namespace',
    dataset_name='test_dataset',
    data_schema=schema(
      {
        'str': 'string',
        'int': 'int32',
        'test_label': field(fields={'label': 'string', 'created': 'timestamp'}, label='test_label'),
      }
    ),
    num_items=3,
  )

  assert list(dataset.select_rows([PATH_WILDCARD])) == [
    {
      'str': 'a',
      'int': 1,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'b',
      'int': 2,
      'test_label.label': None,
      'test_label.created': None,
    },
    {
      'str': 'c',
      'int': 3,
      'test_label.label': None,
      'test_label.created': None,
    },
  ]

  # Make sure we can filter by the label.
  assert list(
    dataset.select_rows([PATH_WILDCARD], filters=[('test_label.label', 'equals', 'true')])
  ) == [
    {
      'str': 'a',
      'int': 1,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
  ]

  assert dataset.get_label_names() == ['test_label']


@freeze_time(TEST_TIME)
def test_add_labels_sort_limit(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  num_labels = dataset.add_labels('test_label', sort_by=['int'], sort_order=SortOrder.ASC, limit=2)
  assert num_labels == 2
  assert dataset.manifest() == DatasetManifest(
    source=TestSource(),
    namespace='test_namespace',
    dataset_name='test_dataset',
    data_schema=schema(
      {
        'str': 'string',
        'int': 'int32',
        'test_label': field(fields={'label': 'string', 'created': 'timestamp'}, label='test_label'),
      }
    ),
    num_items=3,
  )

  assert list(dataset.select_rows([PATH_WILDCARD])) == [
    {
      'str': 'a',
      'int': 1,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'b',
      'int': 2,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'c',
      'int': 3,
      'test_label.label': None,
      'test_label.created': None,
    },
  ]

  assert dataset.get_label_names() == ['test_label']


@freeze_time(TEST_TIME)
def test_add_row_labels(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  num_labels = dataset.add_labels('test_label', row_ids=['1', '2'])
  assert num_labels == 2
  assert list(dataset.select_rows([PATH_WILDCARD])) == [
    {
      'str': 'a',
      'int': 1,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'b',
      'int': 2,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'c',
      'int': 3,
      'test_label.label': None,
      'test_label.created': None,
    },
  ]
  assert dataset.get_label_names() == ['test_label']


@freeze_time(TEST_TIME)
def test_label_deleted_interaction(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  # Expected semantics:
  # Adding/removing labels on deleted rows will fail unless include_deleted=True.
  # That's true even if the row was manually specified in row_ids!
  dataset.add_labels('test_label', row_ids=['1', '2'])
  dataset.delete_rows(['1'])
  dataset.remove_labels('test_label', row_ids=['1', '2'])
  assert list(dataset.select_rows([PATH_WILDCARD], include_deleted=True)) == [
    {
      'str': 'a',
      'int': 1,
      f'{DELETED_LABEL_NAME}.label': 'true',
      f'{DELETED_LABEL_NAME}.created': Timestamp(TEST_TIME),
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'b',
      'int': 2,
      f'{DELETED_LABEL_NAME}.label': None,
      f'{DELETED_LABEL_NAME}.created': None,
      'test_label.label': None,
      'test_label.created': None,
    },
    {
      'str': 'c',
      'int': 3,
      f'{DELETED_LABEL_NAME}.label': None,
      f'{DELETED_LABEL_NAME}.created': None,
      'test_label.label': None,
      'test_label.created': None,
    },
  ]

  dataset.remove_labels('test_label', row_ids=['1'], include_deleted=True)
  assert sorted(
    dataset.select_rows([PATH_WILDCARD], include_deleted=True), key=lambda x: x['int']
  ) == [
    {
      'str': 'a',
      'int': 1,
      f'{DELETED_LABEL_NAME}.label': 'true',
      f'{DELETED_LABEL_NAME}.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'b',
      'int': 2,
      f'{DELETED_LABEL_NAME}.label': None,
      f'{DELETED_LABEL_NAME}.created': None,
    },
    {
      'str': 'c',
      'int': 3,
      f'{DELETED_LABEL_NAME}.label': None,
      f'{DELETED_LABEL_NAME}.created': None,
    },
  ]

  dataset.add_labels('test_label', row_ids=['1', '3'])
  assert sorted(
    dataset.select_rows([PATH_WILDCARD], include_deleted=True), key=lambda x: x['int']
  ) == [
    {
      'str': 'a',
      'int': 1,
      f'{DELETED_LABEL_NAME}.label': 'true',
      f'{DELETED_LABEL_NAME}.created': Timestamp(TEST_TIME),
      'test_label.label': None,
      'test_label.created': None,
    },
    {
      'str': 'b',
      'int': 2,
      f'{DELETED_LABEL_NAME}.label': None,
      f'{DELETED_LABEL_NAME}.created': None,
      'test_label.label': None,
      'test_label.created': None,
    },
    {
      'str': 'c',
      'int': 3,
      f'{DELETED_LABEL_NAME}.label': None,
      f'{DELETED_LABEL_NAME}.created': None,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
  ]


@freeze_time(TEST_TIME)
def test_add_row_labels_no_filters(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  num_labels = dataset.add_labels('test_label')
  assert num_labels == 3
  assert list(dataset.select_rows([PATH_WILDCARD])) == [
    {
      'str': 'a',
      'int': 1,
      'test_label.label': 'true',
      'test_label.created': TEST_TIME,
    },
    {
      'str': 'b',
      'int': 2,
      'test_label.label': 'true',
      'test_label.created': TEST_TIME,
    },
    {
      'str': 'c',
      'int': 3,
      'test_label.label': 'true',
      'test_label.created': TEST_TIME,
    },
  ]

  assert dataset.get_label_names() == ['test_label']


@freeze_time(TEST_TIME)
def test_remove_labels(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  num_labels = dataset.add_labels('test_label', row_ids=['1', '2', '3'])
  assert num_labels == 3

  # Remove the first label.
  num_labels = dataset.remove_labels('test_label', row_ids=['1'])
  assert num_labels == 1

  assert list(dataset.select_rows([PATH_WILDCARD], sort_by=['int'], sort_order=SortOrder.ASC)) == [
    {
      'str': 'a',
      'int': 1,
      'test_label.label': None,
      'test_label.created': None,
    },
    {
      'str': 'b',
      'int': 2,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
    {
      'str': 'c',
      'int': 3,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
  ]

  # Remove by a filter.
  dataset.remove_labels('test_label', filters=[('int', 'greater_equal', 2)])
  assert list(dataset.select_rows([PATH_WILDCARD], sort_by=['int'], sort_order=SortOrder.ASC)) == [
    {'str': 'a', 'int': 1},
    {'str': 'b', 'int': 2},
    {'str': 'c', 'int': 3},
  ]

  assert dataset.get_label_names() == []


@freeze_time(TEST_TIME)
def test_remove_labels_no_filters(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  # Add all labels.
  num_labels = dataset.add_labels('test_label')
  assert num_labels == 3

  # Remove all labels.
  num_labels = dataset.remove_labels('test_label')
  assert num_labels == 3

  assert list(dataset.select_rows([PATH_WILDCARD], sort_by=['int'], sort_order=SortOrder.ASC)) == [
    {'str': 'a', 'int': 1},
    {'str': 'b', 'int': 2},
    {'str': 'c', 'int': 3},
  ]

  assert dataset.get_label_names() == []


@freeze_time(TEST_TIME)
def test_remove_labels_sort_limit(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  # Add labels to every row.
  num_labels = dataset.add_labels('test_label')
  assert num_labels == 3

  # Remove 2 labels.
  num_labels = dataset.remove_labels(
    'test_label', sort_by=['int'], sort_order=SortOrder.ASC, limit=2
  )
  assert num_labels == 2
  assert dataset.manifest() == DatasetManifest(
    source=TestSource(),
    namespace='test_namespace',
    dataset_name='test_dataset',
    data_schema=schema(
      {
        'str': 'string',
        'int': 'int32',
        'test_label': field(fields={'label': 'string', 'created': 'timestamp'}, label='test_label'),
      }
    ),
    num_items=3,
  )

  assert list(dataset.select_rows([PATH_WILDCARD], sort_by=('str',), sort_order=SortOrder.ASC)) == [
    {
      'str': 'a',
      'int': 1,
      'test_label.label': None,
      'test_label.created': None,
    },
    {
      'str': 'b',
      'int': 2,
      'test_label.label': None,
      'test_label.created': None,
    },
    {
      'str': 'c',
      'int': 3,
      'test_label.label': 'true',
      'test_label.created': Timestamp(TEST_TIME),
    },
  ]

  assert dataset.get_label_names() == ['test_label']


@freeze_time(TEST_TIME)
def test_label_overwrites(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  num_labels = dataset.add_labels('test_label', value='yes', filters=[(ROWID, 'equals', '1')])
  assert num_labels == 1
  # Overwrite the value.
  num_labels = dataset.add_labels('test_label', value='no', filters=[(ROWID, 'equals', '1')])
  assert num_labels == 1

  assert list(dataset.select_rows([PATH_WILDCARD])) == [
    {'str': 'a', 'int': 1, 'test_label.label': 'no', 'test_label.created': Timestamp(TEST_TIME)},
    {'str': 'b', 'int': 2, 'test_label.label': None, 'test_label.created': None},
    {'str': 'c', 'int': 3, 'test_label.label': None, 'test_label.created': None},
  ]

  assert dataset.get_label_names() == ['test_label']


@freeze_time(TEST_TIME)
def test_add_multiple_labels(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  # Add a 'yes' for every item with int >= 2.
  num_labels = dataset.add_labels('test_label', value='yes', filters=[('int', 'greater_equal', 2)])
  assert num_labels == 2
  # Add a 'no' for every item with int < 2.
  num_labels = dataset.add_labels('test_label', value='no', filters=[('int', 'less', 2)])
  assert num_labels == 1

  assert dataset.manifest() == DatasetManifest(
    source=TestSource(),
    namespace='test_namespace',
    dataset_name='test_dataset',
    data_schema=schema(
      {
        'str': 'string',
        'int': 'int32',
        'test_label': field(fields={'label': 'string', 'created': 'timestamp'}, label='test_label'),
      }
    ),
    num_items=3,
  )

  assert list(dataset.select_rows([PATH_WILDCARD])) == [
    {'str': 'a', 'int': 1, 'test_label.label': 'no', 'test_label.created': TEST_TIME},
    {'str': 'b', 'int': 2, 'test_label.label': 'yes', 'test_label.created': TEST_TIME},
    {'str': 'c', 'int': 3, 'test_label.label': 'yes', 'test_label.created': TEST_TIME},
  ]

  assert dataset.get_label_names() == ['test_label']


@freeze_time(TEST_TIME)
def test_labels_select_groups(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data(TEST_ITEMS)

  # Add a 'yes' for every item with int >= 2.
  num_labels = dataset.add_labels('test_label', value='yes', filters=[('int', 'greater_equal', 2)])
  assert num_labels == 2

  # Add a 'no' for every item with int < 2.
  num_labels = dataset.add_labels('test_label', value='no', filters=[('int', 'less', 2)])
  assert num_labels == 1

  assert dataset.select_groups(('test_label', 'label')) == SelectGroupsResult(
    too_many_distinct=False, counts=[('yes', 2), ('no', 1)]
  )

  assert dataset.get_label_names() == ['test_label']



"""Test for the concept label signal."""

import os
import pathlib
from typing import Generator, Type

import pytest
from pytest_mock import MockerFixture

from ..concepts.concept import ExampleIn
from ..concepts.db_concept import ConceptDB, ConceptUpdate, DiskConceptDB, DiskConceptModelDB
from ..data.dataset_duckdb import DatasetDuckDB
from ..db_manager import set_default_dataset_cls
from ..schema import SignalInputType, span
from ..signal import clear_signal_registry
from .concept_labels import ConceptLabelsSignal

ALL_CONCEPT_DBS = [DiskConceptDB]
ALL_CONCEPT_MODEL_DBS = [DiskConceptModelDB]


@pytest.fixture(autouse=True)
def set_project_dir(tmp_path: pathlib.Path, mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_PROJECT_DIR': str(tmp_path)})


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Generator:
  # Setup.
  set_default_dataset_cls(DatasetDuckDB)

  # Unit test runs.
  yield

  # Teardown.
  clear_signal_registry()


def test_concept_does_not_exist() -> None:
  signal = ConceptLabelsSignal(namespace='test', concept_name='concept_doesnt_exist')
  with pytest.raises(ValueError, match='Concept "test/concept_doesnt_exist" does not exist'):
    list(signal.compute(['a new data point', 'not in concept']))


@pytest.mark.parametrize('concept_db_cls', ALL_CONCEPT_DBS)
def test_concept_labels(concept_db_cls: Type[ConceptDB]) -> None:
  concept_db = concept_db_cls()
  namespace = 'test'
  concept_name = 'test_concept'
  concept_db.create(namespace=namespace, name=concept_name, type=SignalInputType.TEXT)

  train_data = [
    ExampleIn(label=False, text='no in concept'),
    ExampleIn(label=True, text='yes in concept'),
    # This should never show since we request the main draft.
    ExampleIn(label=False, text='this is unrelated', draft='test_draft'),
  ]
  concept_db.edit(namespace, concept_name, ConceptUpdate(insert=train_data))

  signal = ConceptLabelsSignal(namespace='test', concept_name='test_concept')
  results = list(
    signal.compute(
      [
        'this is no in concept',
        'this is yes in concept',
        'this is no in concept. filler. this is yes in concept.',
        'this is unrelated',
      ]
    )
  )

  assert results == [
    [span(8, 8 + len('no in concept'), {'label': False})],
    [span(8, 8 + len('yes in concept'), {'label': True})],
    [
      span(8, 8 + len('no in concept'), {'label': False}),
      span(39, 39 + len('yes in concept'), {'label': True}),
    ],
    # This example is in the draft, which was not requested.
    None,
  ]


@pytest.mark.parametrize('concept_db_cls', ALL_CONCEPT_DBS)
def test_concept_labels_draft(concept_db_cls: Type[ConceptDB]) -> None:
  concept_db = concept_db_cls()
  namespace = 'test'
  concept_name = 'test_concept'
  concept_db.create(namespace=namespace, name=concept_name, type=SignalInputType.TEXT)

  concept_update = ConceptUpdate(
    insert=[
      ExampleIn(label=True, text='in concept'),
      ExampleIn(label=False, text='out of concept'),
      ExampleIn(label=True, text='in draft', draft='test_draft'),
      ExampleIn(label=False, text='out draft', draft='test_draft'),
    ]
  )

  concept_db.edit(namespace, concept_name, concept_update)

  signal = ConceptLabelsSignal(namespace='test', concept_name='test_concept', draft='test_draft')
  results = list(signal.compute(['this is in concept', 'this is in draft', 'this is out draft']))

  assert results == [
    [span(8, 8 + len('in concept'), {'label': True})],
    [span(8, 8 + len('in draft'), {'label': True, 'draft': 'test_draft'})],
    [span(8, 8 + len('out draft'), {'label': False, 'draft': 'test_draft'})],
  ]


@pytest.mark.parametrize('concept_db_cls', ALL_CONCEPT_DBS)
def test_concept_labels_key(concept_db_cls: Type[ConceptDB]) -> None:
  concept_db = concept_db_cls()
  namespace = 'test'
  concept_name = 'test_concept'
  concept_db.create(namespace=namespace, name=concept_name, type=SignalInputType.TEXT)

  signal = ConceptLabelsSignal(namespace='test', concept_name='test_concept')
  assert signal.key() == 'test/test_concept/labels/preview'


@pytest.mark.parametrize('concept_db_cls', ALL_CONCEPT_DBS)
def test_concept_labels_compute_signal_key(concept_db_cls: Type[ConceptDB]) -> None:
  concept_db = concept_db_cls()
  namespace = 'test'
  concept_name = 'test_concept'
  concept_db.create(namespace=namespace, name=concept_name, type=SignalInputType.TEXT)

  signal = ConceptLabelsSignal(namespace='test', concept_name='test_concept')
  assert signal.key(is_computed_signal=True) == 'test/test_concept/labels'



"""CSV source."""
import os
from typing import ClassVar, Optional, cast

import duckdb
import pyarrow as pa
from pydantic import Field
from typing_extensions import override

from ..data import dataset_utils
from ..schema import PARQUET_FILENAME_PREFIX, ROWID, Schema, arrow_schema_to_schema
from ..source import Source, SourceManifest, SourceSchema
from ..tasks import TaskId
from ..utils import download_http_files
from .duckdb_utils import convert_path_to_duckdb, duckdb_setup

LINE_NUMBER_COLUMN = '__line_number__'


class CSVSource(Source):
  r"""CSV or TSV data loader

  CSV or TSV files can live locally as a filepath, point to an external URL, or live on S3, GCS, or
  R2.

  For TSV files, use `\t` as the delimiter.

  For more details on authorizing access to S3, GCS or R2, see:
  https://duckdb.org/docs/guides/import/s3_import.html
  """  # noqa: D415, D400

  name: ClassVar[str] = 'csv'

  filepaths: list[str] = Field(
    description='A list of paths to CSV files. '
    'Paths can be local, point to an HTTP(s) url, or live on GCS, S3 or R2.'
  )
  delim: Optional[str] = Field(
    default=',', description='The CSV file delimiter to use. For TSV files, use `\\t`.'
  )
  header: Optional[bool] = Field(default=True, description='Whether the CSV file has a header row.')
  names: list[str] = Field(
    default=[], description='Provide header names if the file does not contain a header.'
  )

  _source_schema: Optional[SourceSchema] = None
  _reader: Optional[pa.RecordBatchReader] = None
  _con: Optional[duckdb.DuckDBPyConnection] = None

  @override
  def setup(self) -> None:
    # Download CSV files to /tmp if they are via HTTP to speed up duckdb.
    filepaths = download_http_files(self.filepaths)

    self._con = duckdb.connect(database=':memory:')
    duckdb_setup(self._con)

    # DuckDB expects s3 protocol: https://duckdb.org/docs/guides/import/s3_import.html.
    duckdb_paths = [convert_path_to_duckdb(path) for path in filepaths]

    # NOTE: We use duckdb here to increase parallelism for multiple files.
    # NOTE: We turn off the parallel reader because of https://github.com/lilacai/lilac/issues/373.
    self._con.execute(
      f"""
      CREATE SEQUENCE serial START 1;
      CREATE VIEW t as (SELECT nextval('serial') as "{LINE_NUMBER_COLUMN}", * FROM read_csv_auto(
        {duckdb_paths},
        SAMPLE_SIZE=500000,
        HEADER={self.header},
        {f'NAMES={self.names},' if self.names else ''}
        DELIM='{self.delim or ','}',
        IGNORE_ERRORS=true,
        PARALLEL=false
    ));
    """
    )

    res = self._con.execute('SELECT COUNT(*) FROM t').fetchone()
    num_items = cast(tuple[int], res)[0]

    self._reader = self._con.execute('SELECT * from t').fetch_record_batch(rows_per_batch=10_000)
    # Create the source schema in prepare to share it between process and source_schema.
    schema = arrow_schema_to_schema(self._reader.schema)
    self._source_schema = SourceSchema(fields=schema.fields, num_items=num_items)

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    assert self._source_schema is not None
    return self._source_schema

  @override
  def load_to_parquet(self, output_dir: str, task_id: Optional[TaskId]) -> SourceManifest:
    del task_id

    assert self._con, 'setup() must be called first.'

    out_filename = dataset_utils.get_parquet_filename(PARQUET_FILENAME_PREFIX, 0, 1)
    filepath = os.path.join(output_dir, out_filename)
    os.makedirs(output_dir, exist_ok=True)

    self._con.sql(
      f"SELECT replace(CAST(uuid() AS VARCHAR), '-', '') AS {ROWID}, * FROM t"
    ).write_parquet(filepath, compression='zstd')
    schema = Schema(fields=self.source_schema().fields.copy())
    return SourceManifest(files=[out_filename], data_schema=schema, source=self)



"""Interface for implementing a signal."""

import abc
import copy
from typing import (
  Any,
  Callable,
  ClassVar,
  Iterable,
  Iterator,
  Literal,
  Optional,
  Sequence,
  Type,
  TypeVar,
  Union,
)

from pydantic import BaseModel, ConfigDict, model_serializer
from pydantic import Field as PydanticField
from typing_extensions import override

from .embeddings.vector_store import VectorDBIndex
from .schema import (
  EMBEDDING_KEY,
  EmbeddingInputType,
  Field,
  Item,
  PathKey,
  SignalInputType,
  SpanVector,
  field,
)
from .tasks import TaskExecutionType


def _signal_schema_extra(schema: dict[str, Any], signal: Type['Signal']) -> None:
  """Add the title to the schema from the display name and name.

  Pydantic defaults this to the class name.
  """
  if hasattr(signal, 'display_name'):
    schema['title'] = signal.display_name
  schema['runs_remote'] = signal.runs_remote
  if not signal.runs_remote and 'remote' in schema['properties']:
    del schema['properties']['remote']

  signal_prop: dict[str, Any]
  if hasattr(signal, 'name'):
    signal_prop = {'enum': [signal.name]}
  else:
    signal_prop = {'type': 'string'}
  schema['properties'] = {'signal_name': signal_prop, **schema['properties']}
  if 'required' not in schema:
    schema['required'] = []
  schema['required'].append('signal_name')


OutputType = Optional[Literal['embedding', 'cluster']]
# A function that takes a list of (topic, membership_score) tuples and returns a single topic.
TopicFn = Callable[[list[tuple[str, float]]], str]


class Signal(BaseModel):
  """Interface for signals to implement. A signal can score documents and a dataset column."""

  # ClassVars do not get serialized with pydantic.
  name: ClassVar[str]
  # The display name is just used for rendering in the UI.
  display_name: ClassVar[Optional[str]]

  # The input type is used to populate the UI to determine what the signal accepts as input.
  input_type: ClassVar[SignalInputType]

  # The output type is the logical type, and treated in a special way in the UI.
  output_type: ClassVar[OutputType] = None

  # See lilac.data.dataset.Dataset.map for definitions and semantics.
  local_batch_size: ClassVar[Optional[int]] = -1
  local_parallelism: ClassVar[int] = 1
  local_strategy: ClassVar[TaskExecutionType] = 'threads'

  runs_remote: ClassVar[bool] = False

  remote: bool = PydanticField(default=False, description='Accelerate computation on Lilac Garden')

  @model_serializer(mode='wrap', when_used='always')
  def serialize_model(self, serializer: Callable[..., dict[str, Any]]) -> dict[str, Any]:
    """Serialize the model to a dictionary."""
    res = serializer(self)
    res['signal_name'] = self.name
    return res

  model_config = ConfigDict(extra='ignore', json_schema_extra=_signal_schema_extra)

  def fields(self) -> Optional[Field]:
    """Return the schema for the signal output. If None, the schema will be automatically inferred.

    Returns:
      A Field object that describes the schema of the signal.
    """
    return None

  def compute(self, data: Any) -> Any:
    """Compute a signal over a field.

    Just like in dataset.map, if the field is a repeated field, then the repeated values will be
    flattened and presented as a continuous stream.

    A signal can choose to return None for any input, but it must always return some value, so that
    alignment errors don't occur when Lilac attempts to unflatten repeated fields.

    This function is polymorphic and its signature depends on the local_batch_size class var.

    If batch_size = -1 (default), then we hand you the iterator stream and you choose how to process
    the stream; the function signature is:
      compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]
    If batch_size > 0, then we will batch items and provide them to you in the requested batch size;
    the function signature is:
      compute(self, data: list[RichData]) -> list[Optional[Item]]
    If batch_size is None, then we hand you items one by one, and the function signature is:
      compute(self, data: RichData) -> Optional[Item]

    Args:
      data: An iterable of rich data to compute the signal over.
      user: User information, if the user is logged in. This is useful if signals are access
        controlled, like concepts.

    Returns:
      An iterable of items. Sparse signals should return "None" for skipped inputs.
    """
    raise NotImplementedError

  def compute_remote(self, data: Iterator[Any]) -> Iterator[Any]:
    """Compute a signal over a field, but on a remote machine.

    This method gets an iterator of the entire data, and should return an iterator of the same
    length, with the processed results.
    """
    raise NotImplementedError

  def key(self, is_computed_signal: Optional[bool] = False) -> str:
    """Get the key for a signal.

    This is used to make sure signals with multiple arguments do not collide.

    NOTE: Overriding this method is sensitive. If you override it, make sure that it is globally
    unique. It will be used as the dictionary key for enriched values.

    Args:
      is_computed_signal: True when the signal is computed over the column and written to
        disk. False when the signal is used as a preview UDF.
    """
    args_dict = self.model_dump(exclude_unset=True, exclude_defaults=True)
    # If a user explicitly defines a signal name for whatever reason, remove it as it's redundant.
    if 'signal_name' in args_dict:
      del args_dict['signal_name']

    return self.name + _args_key_from_dict(args_dict)

  def setup(self) -> None:
    """Setup the signal."""
    pass

  def teardown(self) -> None:
    """Tears down the signal."""
    pass

  def __str__(self) -> str:
    return f' {self.__class__.__name__}({self.model_dump_json(exclude_none=True)})'


def _args_key_from_dict(args_dict: dict[str, Any]) -> str:
  args = None
  args_list: list[str] = []
  for k, v in args_dict.items():
    if v:
      args_list.append(f'{k}={v}')

  args = ','.join(args_list)
  return '' if not args_list else f'({args})'


# Signal base classes, used for inferring the dependency chain required for computing a signal.
class TextSignal(Signal):
  """An interface for signals that compute over text."""

  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  @override
  def key(self, is_computed_signal: Optional[bool] = False) -> str:
    args_dict = self.model_dump(exclude_unset=True, exclude_defaults=True)
    if 'signal_name' in args_dict:
      del args_dict['signal_name']
    return self.name + _args_key_from_dict(args_dict)


class TextEmbeddingSignal(TextSignal):
  """An interface for signals that compute embeddings for text."""

  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT
  embed_input_type: EmbeddingInputType = PydanticField(
    title='Embedding Input Type', default='document', description='The input type to the embedding.'
  )
  output_type: ClassVar[OutputType] = 'embedding'

  _split = True

  def __init__(self, split: bool = True, **kwargs: Any):
    super().__init__(**kwargs)
    self._split = split

  @override
  def fields(self) -> Field:
    """NOTE: Override this method at your own risk if you want to add extra metadata.

    Embeddings should not come with extra metadata.
    """
    return field(fields=[field('string_span', fields={EMBEDDING_KEY: 'embedding'})])


def _vector_signal_schema_extra(schema: dict[str, Any], signal: Type['Signal']) -> None:
  """Add the enum values for embeddings."""
  embeddings: list[str] = []
  for s in SIGNAL_REGISTRY.values():
    if issubclass(s, TextEmbeddingSignal):
      embeddings.append(s.name)

  if hasattr(signal, 'display_name'):
    schema['title'] = signal.display_name

  signal_prop: dict[str, Any]
  if hasattr(signal, 'name'):
    signal_prop = {'enum': [signal.name]}
  else:
    signal_prop = {'type': 'string'}
  schema['properties'] = {'signal_name': signal_prop, **schema['properties']}
  if 'required' not in schema:
    schema['required'] = []
  schema['required'].append('signal_name')

  schema['properties']['embedding']['enum'] = embeddings


class VectorSignal(Signal, abc.ABC):
  """An interface for signals that can compute items given vector inputs."""

  embedding: str = PydanticField(description='The name of the pre-computed embedding.')

  model_config = ConfigDict(json_schema_extra=_vector_signal_schema_extra)

  @abc.abstractmethod
  def vector_compute(self, vector_spans: Iterable[list[SpanVector]]) -> Iterator[Optional[Item]]:
    """Compute the signal for an iterable of keys that point to documents or images.

    Args:
      vector_spans: Precomputed embeddings over spans of a document.

    Returns:
      An iterable of items. Sparse signals should return "None" for skipped inputs.
    """
    raise NotImplementedError

  def vector_compute_topk(
    self, topk: int, vector_index: VectorDBIndex, rowids: Optional[Iterable[str]] = None
  ) -> Sequence[tuple[PathKey, Optional[Item]]]:
    """Return signal results only for the top k documents or images.

    Signals decide how to rank each document/image in the dataset, usually by a similarity score
    obtained via the vector store.

    Args:
      topk: The number of items to return, ranked by the signal.
      vector_index: The vector index to lookup pre-computed embeddings.
      rowids: Optional iterable of row ids to restrict the search to.

    Returns:
      A list of (key, signal_output) tuples containing the `topk` items. Sparse signals should
      return "None" for skipped inputs.
    """
    raise NotImplementedError


Tsignal = TypeVar('Tsignal', bound=Signal)


def get_signal_by_type(signal_name: str, signal_type: Type[Tsignal]) -> Type[Tsignal]:
  """Return a signal class by name and signal type."""
  if signal_name not in SIGNAL_REGISTRY:
    raise ValueError(f'Signal "{signal_name}" not found in the registry')

  signal_cls = SIGNAL_REGISTRY[signal_name]
  if not issubclass(signal_cls, signal_type):
    raise ValueError(
      f'"{signal_name}" is a `{signal_cls.__name__}`, '
      f'which is not a subclass of `{signal_type.__name__}`.'
    )
  return signal_cls


def get_signals_by_type(signal_type: Type[Tsignal]) -> list[Type[Tsignal]]:
  """Return all signals that match a signal type."""
  signal_clses: list[Type[Tsignal]] = []
  for signal_cls in SIGNAL_REGISTRY.values():
    if issubclass(signal_cls, signal_type):
      signal_clses.append(signal_cls)
  return signal_clses


SIGNAL_REGISTRY: dict[str, Type[Signal]] = {}


def register_signal(signal_cls: Type[Signal]) -> None:
  """Register a signal in the global registry."""
  if signal_cls.name in SIGNAL_REGISTRY:
    raise ValueError(f'Signal "{signal_cls.name}" has already been registered!')

  SIGNAL_REGISTRY[signal_cls.name] = signal_cls


def get_signal_cls(signal_name: str) -> Optional[Type[Signal]]:
  """Return a registered signal given the name in the registry."""
  return SIGNAL_REGISTRY.get(signal_name)


def resolve_signal(signal: Union[dict, Signal]) -> Signal:
  """Resolve a generic signal base class to a specific signal class."""
  if isinstance(signal, Signal):
    # The signal config is already parsed.
    return signal

  signal_name = signal.get('signal_name')
  if not signal_name:
    raise ValueError('"signal_name" needs to be defined in the json dict.')

  signal_cls = get_signal_cls(signal_name)
  if not signal_cls:
    # Make a metaclass so we get a valid `Signal` class.
    signal_cls = type(f'Signal_{signal_name}', (Signal,), {'name': signal_name})

  signal = copy.deepcopy(signal)
  del signal['signal_name']
  return signal_cls(**signal)


def clear_signal_registry() -> None:
  """Clear the signal registry."""
  SIGNAL_REGISTRY.clear()



"""Routing endpoints for running signals on datasets."""
from threading import Thread
from typing import Annotated, Optional

from fastapi import APIRouter, HTTPException
from fastapi.params import Depends
from pydantic import BaseModel, SerializeAsAny, field_validator

from .auth import UserInfo, get_session_user, get_user_access
from .db_manager import get_dataset
from .router_utils import RouteErrorHandler
from .schema import Path
from .signal import Signal, TextEmbeddingSignal, resolve_signal
from .tasks import TaskId, get_task_manager, launch_task

router = APIRouter(route_class=RouteErrorHandler)


class ComputeSignalOptions(BaseModel):
  """The request for the compute signal endpoint."""

  signal: SerializeAsAny[Signal]

  # The leaf path to compute the signal on.
  leaf_path: Path

  @field_validator('signal', mode='before')
  @classmethod
  def parse_signal(cls, signal: dict) -> Signal:
    """Parse a signal to its specific subclass instance."""
    return resolve_signal(signal)


class ComputeSignalResponse(BaseModel):
  """Response of the compute signal column endpoint."""

  task_id: TaskId


@router.post('/{namespace}/{dataset_name}/compute_signal')
def compute_signal(
  namespace: str,
  dataset_name: str,
  options: ComputeSignalOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> ComputeSignalResponse:
  """Compute a signal for a dataset."""
  if not get_user_access(user).dataset.compute_signals:
    raise HTTPException(401, 'User does not have access to compute signals over this dataset.')

  # Resolve the signal outside the task so we don't look up the signal in the registry. This gets
  # implicitly pickled by the serializer when _task_compute_signal is pickled.
  # NOTE: This unfortunately does not work in Jupyter because a module is not picklable. In this
  # case, we recommend defining and registering the signal outside a Jupyter notebook.
  signal = options.signal

  path_str = '.'.join(map(str, options.leaf_path))
  task_id = get_task_manager().task_id(
    name=f'[{namespace}/{dataset_name}] Compute signal "{options.signal.name}" on "{path_str}"',
    description=f'Config: {options.signal}',
  )

  dataset = get_dataset(namespace, dataset_name)

  def run() -> None:
    dataset.compute_signal(
      signal,
      options.leaf_path,
      # Overwrite for text embeddings since we don't have UI to control deleting embeddings.
      overwrite=isinstance(options.signal, TextEmbeddingSignal),
      task_id=task_id,
      remote=signal.remote,
    )

  launch_task(task_id, run)

  return ComputeSignalResponse(task_id=task_id)


class ClusterOptions(BaseModel):
  """The request for the cluster endpoint."""

  input: Path
  output_path: Optional[Path] = None
  remote: bool = False
  overwrite: bool = False


class ClusterResponse(BaseModel):
  """Response of the cluster endpoint."""

  task_id: TaskId


@router.post('/{namespace}/{dataset_name}/cluster')
def cluster(
  namespace: str,
  dataset_name: str,
  options: ClusterOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> ClusterResponse:
  """Compute clusters over an input path."""
  if not get_user_access(user).dataset.compute_signals:
    raise HTTPException(401, 'User does not have access to compute clusters over this dataset.')

  path_str = '.'.join(map(str, options.input))
  task_name = f'[{namespace}/{dataset_name}] Clustering "{path_str}"'
  task_id = get_task_manager().task_id(name=task_name)
  dataset = get_dataset(namespace, dataset_name)

  def run() -> None:
    dataset.cluster(
      options.input,
      options.output_path,
      remote=options.remote,
      overwrite=options.overwrite,
      task_id=task_id,
    )

  thread = Thread(target=run, daemon=True)
  thread.start()
  return ClusterResponse(task_id=task_id)


class DeleteSignalOptions(BaseModel):
  """The request for the delete signal endpoint."""

  # The signal path holding the data from the signal.
  signal_path: Path


class DeleteSignalResponse(BaseModel):
  """Response of the compute signal column endpoint."""

  completed: bool


@router.delete('/{namespace}/{dataset_name}/delete_signal')
def delete_signal(
  namespace: str,
  dataset_name: str,
  options: DeleteSignalOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> DeleteSignalResponse:
  """Delete a signal from a dataset."""
  if not get_user_access(user).dataset.delete_signals:
    raise HTTPException(401, 'User does not have access to delete this signal.')

  dataset = get_dataset(namespace, dataset_name)
  dataset.delete_signal(options.signal_path)
  return DeleteSignalResponse(completed=True)



"""Tests for load.py: loading project configs."""

import os
import pathlib
import re
from typing import ClassVar, Iterable, Iterator, Optional, cast

import numpy as np
import pytest
from pytest_mock import MockerFixture
from typing_extensions import override

from .config import (
  ClusterConfig,
  ClusterInputSelectorConfig,
  Config,
  DatasetConfig,
  EmbeddingConfig,
  SignalConfig,
)
from .data import clustering
from .data.dataset import DatasetManifest
from .db_manager import get_dataset
from .env import set_project_dir
from .load import load
from .project import PROJECT_CONFIG_FILENAME, init
from .schema import EMBEDDING_KEY, Field, Item, RichData, field, lilac_embedding, schema
from .signal import TextEmbeddingSignal, TextSignal, clear_signal_registry, register_signal
from .source import Source, SourceSchema, clear_source_registry, register_source
from .utils import to_yaml

SIMPLE_ITEMS: list[Item] = [
  {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  {'str': 'c', 'int': 3, 'bool': True, 'float': 1.0},
]

EMBEDDINGS: list[tuple[str, list[float]]] = [
  ('a', [1.0, 0.0, 0.0]),
  ('b', [1.0, 1.0, 0.0]),
  ('c', [1.0, 1.0, 0.0]),
]

STR_EMBEDDINGS: dict[str, list[float]] = {text: embedding for text, embedding in EMBEDDINGS}


class TestSource(Source):
  """A test source."""

  name: ClassVar[str] = 'test_source'

  @override
  def source_schema(self) -> SourceSchema:
    """Yield all items."""
    return SourceSchema(
      fields={
        'str': field('string'),
        'int': field('int32'),
        'bool': field('boolean'),
        'float': field('float32'),
      },
      num_items=len(SIMPLE_ITEMS),
    )

  @override
  def yield_items(self) -> Iterable[Item]:
    """Yield all items."""
    yield from SIMPLE_ITEMS


class TestEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      yield [lilac_embedding(0, len(example), np.array(STR_EMBEDDINGS[cast(str, example)]))]


class TestSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  _call_count: int = 0

  def fields(self) -> Field:
    return field('int32')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      self._call_count += 1
      yield len(text_content)


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_source(TestSource)
  register_signal(TestSignal)
  register_signal(TestEmbedding)

  # Unit test runs.
  yield

  # Teardown.
  clear_source_registry()
  clear_signal_registry()


def test_load_config_obj(tmp_path: pathlib.Path) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  project_config = Config(
    datasets=[DatasetConfig(namespace='namespace', name='test', source=TestSource())]
  )

  # Load the project config from a config object.
  load(config=project_config)

  dataset = get_dataset('namespace', 'test')

  assert dataset.manifest() == DatasetManifest(
    namespace='namespace',
    dataset_name='test',
    data_schema=schema({'str': 'string', 'int': 'int32', 'bool': 'boolean', 'float': 'float32'}),
    num_items=3,
    source=TestSource(),
  )


def test_load_project_config_yml(tmp_path: pathlib.Path) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  # Simulate the user manually editing the project config.
  project_config = Config(
    datasets=[DatasetConfig(namespace='namespace', name='test', source=TestSource())]
  )
  project_config_yml = to_yaml(project_config.model_dump())
  config_path = os.path.join(tmp_path, PROJECT_CONFIG_FILENAME)

  # Write the project_config as yml to project dir.
  with open(config_path, 'w') as f:
    f.write(project_config_yml)

  load(config=config_path)

  dataset = get_dataset('namespace', 'test')

  assert dataset.manifest() == DatasetManifest(
    namespace='namespace',
    dataset_name='test',
    data_schema=schema({'str': 'string', 'int': 'int32', 'bool': 'boolean', 'float': 'float32'}),
    num_items=3,
    source=TestSource(),
  )


def test_load_config_yml_outside_project(tmp_path: pathlib.Path) -> None:
  # This test makes sure that we can load a yml file from outside the project directory.

  config_path = os.path.join(tmp_path, PROJECT_CONFIG_FILENAME)
  project_dir = os.path.join(tmp_path, 'project')
  os.makedirs(project_dir)
  set_project_dir(tmp_path)

  project_config = Config(
    datasets=[DatasetConfig(namespace='namespace', name='test', source=TestSource())]
  )

  project_config_yml = to_yaml(project_config.model_dump())

  # Write the project_config as yml to project dir.
  with open(config_path, 'w') as f:
    f.write(project_config_yml)

  load(config=config_path)

  dataset = get_dataset('namespace', 'test')

  assert dataset.manifest() == DatasetManifest(
    namespace='namespace',
    dataset_name='test',
    data_schema=schema({'str': 'string', 'int': 'int32', 'bool': 'boolean', 'float': 'float32'}),
    num_items=3,
    source=TestSource(),
  )


def test_load_signals(tmp_path: pathlib.Path) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  test_signal = TestSignal()
  project_config = Config(
    datasets=[
      DatasetConfig(
        namespace='namespace',
        name='test',
        source=TestSource(),
        signals=[SignalConfig(path=('str',), signal=test_signal)],
      )
    ]
  )

  # Load the project config from a config object.
  load(config=project_config)

  dataset = get_dataset('namespace', 'test')

  assert dataset.manifest() == DatasetManifest(
    namespace='namespace',
    dataset_name='test',
    data_schema=schema(
      {
        'str': field(
          'string',
          fields={'test_signal': field('int32', signal=test_signal.model_dump(exclude_none=True))},
        ),
        'int': 'int32',
        'bool': 'boolean',
        'float': 'float32',
      }
    ),
    num_items=3,
    source=TestSource(),
  )


def test_load_embeddings(tmp_path: pathlib.Path) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  project_config = Config(
    datasets=[
      DatasetConfig(
        namespace='namespace',
        name='test',
        source=TestSource(),
        embeddings=[EmbeddingConfig(path=('str',), embedding='test_embedding')],
      )
    ]
  )

  # Load the project config from a config object.
  load(config=project_config)

  dataset = get_dataset('namespace', 'test')

  assert dataset.manifest() == DatasetManifest(
    namespace='namespace',
    dataset_name='test',
    data_schema=schema(
      {
        'str': field(
          'string',
          fields={
            'test_embedding': field(
              signal=TestEmbedding().model_dump(exclude_none=True),
              fields=[field('string_span', fields={EMBEDDING_KEY: 'embedding'})],
            )
          },
        ),
        'int': 'int32',
        'bool': 'boolean',
        'float': 'float32',
      }
    ),
    num_items=3,
    source=TestSource(),
  )


def test_load_twice_no_overwrite(tmp_path: pathlib.Path, capsys: pytest.CaptureFixture) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  test_signal = TestSignal()
  project_config = Config(
    datasets=[
      DatasetConfig(
        namespace='namespace',
        name='test',
        source=TestSource(),
        signals=[SignalConfig(path=('str',), signal=test_signal)],
        embeddings=[EmbeddingConfig(path=('str',), embedding='test_embedding')],
      )
    ]
  )

  # Load the project config from a config object.
  load(config=project_config)

  assert 'Computing signal' in capsys.readouterr().out

  first_manifest = get_dataset('namespace', 'test').manifest()

  # Load the project again, make sure signals and embeddings are not computed again.
  load(config=project_config)

  second_manifest = get_dataset('namespace', 'test').manifest()
  assert first_manifest == second_manifest
  assert re.search('Signal  TestSignal(.*) already exists', capsys.readouterr().out)


def test_load_twice_overwrite(tmp_path: pathlib.Path, capsys: pytest.CaptureFixture) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  test_signal = TestSignal()
  project_config = Config(
    datasets=[
      DatasetConfig(
        namespace='namespace',
        name='test',
        source=TestSource(),
        signals=[SignalConfig(path=('str',), signal=test_signal)],
        embeddings=[EmbeddingConfig(path=('str',), embedding='test_embedding')],
      )
    ]
  )

  # Load the project config from a config object.
  load(config=project_config)
  assert 'Computing signal' in capsys.readouterr().out

  first_manifest = get_dataset('namespace', 'test').manifest()

  # Load the project again, make sure signals and embeddings are not computed again.
  load(config=project_config, overwrite=True)

  assert (
    'Signal  TestSignal({"signal_name":"test_signal"}) already exists'
    not in capsys.readouterr().out
  )

  second_manifest = get_dataset('namespace', 'test').manifest()

  assert first_manifest == second_manifest


def test_load_clusters(tmp_path: pathlib.Path, capsys: pytest.CaptureFixture) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  project_config = Config(
    datasets=[
      DatasetConfig(
        namespace='namespace',
        name='test',
        source=TestSource(),
      )
    ],
    clusters=[
      ClusterConfig(
        dataset_namespace='namespace', dataset_name='test', input_path=('str',), remote=False
      )
    ],
  )

  # Load the project config from a config object.
  load(config=project_config)
  assert 'Computing cluster:' in capsys.readouterr().out

  dataset = get_dataset('namespace', 'test')
  assert dataset.manifest().data_schema.fields['str__cluster'].cluster is not None

  load(config=project_config)
  assert 'Cluster already computed:' in capsys.readouterr().out


class TestShareGPTSource(Source):
  """A test sharegpt source."""

  name: ClassVar[str] = 'test_sharegpt_source'

  @override
  def source_schema(self) -> SourceSchema:
    """Yield all items."""
    return SourceSchema(
      fields=schema(
        {
          'conversations': [
            {
              'from': 'string',
              'value': 'string',
            }
          ]
        }
      ).fields,
      num_items=4,
    )

  @override
  def yield_items(self) -> Iterable[Item]:
    """Yield all items."""
    yield from [
      {
        'conversations': [
          {'from': 'human', 'value': 'hello'},
          {'from': 'gpt', 'value': 'i am a language model'},
        ]
      },
      {
        'conversations': [
          {'from': 'human', 'value': 'whats the time'},
          {'from': 'gpt', 'value': '1030'},
        ]
      },
      {
        'conversations': [
          {'from': 'human', 'value': 'hello how are you'},
          {'from': 'gpt', 'value': 'pretty good today'},
        ]
      },
      {
        'conversations': [
          {'from': 'human', 'value': 'whats the hour'},
          {'from': 'gpt', 'value': '10 is the hour'},
        ]
      },
    ]


def test_load_clusters_format_selector(
  tmp_path: pathlib.Path, capsys: pytest.CaptureFixture, mocker: MockerFixture
) -> None:
  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')

  topic_fn_calls: list[list[tuple[str, float]]] = []

  def _test_topic_fn(docs: list[tuple[str, float]]) -> str:
    topic_fn_calls.append(docs)
    if 'hello' in docs[0][0]:
      return 'greeting'
    elif 'time' in docs[0][0] or 'hour' in docs[0][0]:
      return 'time'
    return 'other'

  mocker.patch.object(clustering, 'summarize_request', side_effect=_test_topic_fn)
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  project_config = Config(
    datasets=[
      DatasetConfig(
        namespace='namespace',
        name='test',
        source=TestShareGPTSource(),
      )
    ],
    clusters=[
      ClusterConfig(
        dataset_namespace='namespace',
        dataset_name='test',
        input_selector=ClusterInputSelectorConfig(
          format='sharegpt',
          selector='human',
        ),
        output_path=('cluster',),
        min_cluster_size=2,
        remote=False,
      )
    ],
  )

  # Load the project config from a config object.
  load(config=project_config)
  assert 'Computing cluster:' in capsys.readouterr().out

  # Sort because topics are shuffled.
  for topic_fn_call in topic_fn_calls:
    topic_fn_call.sort()

  # Make sure the topic function is only called for the human text.
  assert topic_fn_calls == [
    [('hello', 1.0), ('hello how are you', 1.0)],
    [('whats the hour', 1.0), ('whats the time', 1.0)],
  ]

  load(config=project_config)
  assert 'Cluster already computed:' in capsys.readouterr().out



"""Tests the vector store interface."""

import pathlib
from typing import Type, cast

import numpy as np
import pytest
from sklearn.preprocessing import normalize

from .vector_store import VectorDBIndex, VectorStore
from .vector_store_hnsw import HNSWVectorStore
from .vector_store_numpy import NumpyVectorStore

ALL_STORES = [NumpyVectorStore, HNSWVectorStore]


@pytest.mark.parametrize('store_cls', ALL_STORES)
class VectorStoreImplSuite:
  def test_add_chunks(self, store_cls: Type[VectorStore]) -> None:
    store = store_cls()

    store.add([('a',), ('b',)], np.array([[1, 2], [3, 4]]))
    store.add([('c',)], np.array([[5, 6]]))

    np.testing.assert_array_equal(
      store.get([('a',), ('b',), ('c',)]), np.array([[1, 2], [3, 4], [5, 6]])
    )

  def test_get_all(self, store_cls: Type[VectorStore]) -> None:
    store = store_cls()

    store.add([('a',), ('b',), ('c',)], np.array([[1, 2], [3, 4], [5, 6]]))

    np.testing.assert_array_equal(
      store.get([('a',), ('b',), ('c',)]), np.array([[1, 2], [3, 4], [5, 6]])
    )

  def test_get_subset(self, store_cls: Type[VectorStore]) -> None:
    store = store_cls()

    store.add([('a',), ('b',), ('c',)], np.array([[1, 2], [3, 4], [5, 6]]))

    np.testing.assert_array_equal(store.get([('b',), ('c',)]), np.array([[3, 4], [5, 6]]))

  def test_save_load(self, store_cls: Type[VectorStore], tmp_path: pathlib.Path) -> None:
    store = store_cls()

    store.add([('a',), ('b',), ('c',)], np.array([[1, 2], [3, 4], [5, 6]]))

    store.save(str(tmp_path))

    del store

    store = store_cls()
    store.load((str(tmp_path)))

    np.testing.assert_array_equal(
      store.get([('a',), ('b',), ('c',)]), np.array([[1, 2], [3, 4], [5, 6]])
    )

  def test_topk(self, store_cls: Type[VectorStore]) -> None:
    store = store_cls()
    embedding = cast(np.ndarray, normalize(np.array([[1, 0], [0, 1], [1, 1]])))
    query = np.array([0.9, 1])
    query /= np.linalg.norm(query)
    topk = 3
    store.add([('a',), ('b',), ('c',)], embedding)
    result = store.topk(query, topk)
    assert [key for key, _ in result] == [('c',), ('b',), ('a',)]
    assert [score for _, score in result] == pytest.approx([0.999, 0.743, 0.669], abs=1e-3)

  def test_topk_with_restricted_keys(self, store_cls: Type[VectorStore]) -> None:
    store = store_cls()
    embedding = np.array([[0.45, 0.89], [0.6, 0.8], [0.64, 0.77]])
    query = np.array([0.89, 0.45])
    topk = 3
    store.add([('a',), ('b',), ('c',)], embedding)
    result = store.topk(query, topk, keys=[('b',), ('a',)])
    assert [key for key, _ in result] == [('b',), ('a',)]
    assert [score for _, score in result] == pytest.approx([0.894, 0.801], 1e-3)

    result = store.topk(query, topk, keys=[('a',), ('b',)])
    assert [key for key, _ in result] == [('b',), ('a',)]
    assert [score for _, score in result] == pytest.approx([0.894, 0.801], 1e-3)

    result = store.topk(query, topk, keys=[('a',), ('c',)])
    assert [key for key, _ in result] == [('c',), ('a',)]
    assert [score for _, score in result] == pytest.approx([0.9161, 0.801], 1e-3)

  def test_topk_with_keys(self, store_cls: Type[VectorStore]) -> None:
    store = store_cls()
    embedding = np.array([[8], [9], [3], [10]])
    store.add([('a', 0), ('a', 1), ('b', 0), ('c', 0)], embedding)
    query = np.array([1])
    result = store.topk(query, k=2, keys=[('b', 0), ('c', 0)])
    assert result == [(('c', 0), 10.0), (('b', 0), 3.0)]

    result = store.topk(query, k=10, keys=[('b', 0), ('a', 1), ('a', 0)])
    assert result == [(('a', 1), 9.0), (('a', 0), 8.0), (('b', 0), 3.0)]


class VectorStoreWrapperSuite:
  def test_topk_with_missing_keys(self) -> None:
    store = VectorDBIndex('numpy')
    all_spans = [
      (('a',), [(0, 1), (2, 3)]),
      (('b',), [(0, 1), (2, 3)]),
      (('c',), [(0, 1), (2, 3)]),
    ]
    embedding = np.array([[8], [9], [3], [10], [11], [12]])
    store.add(all_spans, embedding)

    query = np.array([1])
    result = store.topk(query, k=2, rowids=['a', 'b', 'c', 'd'])
    assert result == [(('c',), 12.0), (('b',), 10.0)]



"""Tests for server.py."""

import os
import pathlib

import pytest

from .config import Config
from .env import get_project_dir, set_project_dir
from .project import PROJECT_CONFIG_FILENAME, create_project_and_set_env, init, read_project_config


async def test_create_project_and_set_env(tmp_path_factory: pytest.TempPathFactory) -> None:
  tmp_path = str(tmp_path_factory.mktemp('test_project'))

  # Test that it creates a project if it doesn't exist.
  create_project_and_set_env(tmp_path)

  assert os.path.exists(os.path.join(tmp_path, PROJECT_CONFIG_FILENAME))
  assert get_project_dir() == tmp_path


async def test_create_project_and_set_env_from_env(
  tmp_path_factory: pytest.TempPathFactory
) -> None:
  tmp_path = str(tmp_path_factory.mktemp('test_project'))

  os.environ['LILAC_PROJECT_DIR'] = tmp_path

  # Test that an empty project directory defaults to the data path.
  create_project_and_set_env(project_dir_arg='')

  assert os.path.exists(os.path.join(tmp_path, PROJECT_CONFIG_FILENAME))
  assert get_project_dir() == tmp_path


def test_init(tmp_path: pathlib.Path) -> None:
  set_project_dir(tmp_path)

  # Initialize the lilac project. init() defaults to the project directory.
  init()

  assert read_project_config(tmp_path) == Config(datasets=[])



"""Deploy to a huggingface space.

Usage:
  poetry run lilac deploy-project

"""
import os
import subprocess
import tempfile
from importlib import resources
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from .concepts.db_concept import CONCEPTS_DIR, DiskConceptDB, get_concept_output_dir
from .config import Config
from .data.dataset_storage_utils import upload
from .env import get_project_dir
from .project import PROJECT_CONFIG_FILENAME, read_project_config, write_project_config
from .utils import get_hf_dataset_repo_id, get_lilac_cache_dir, log, to_yaml

HF_SPACE_DIR = '.hf_spaces'
PY_DIST_DIR = 'dist'
REMOTE_DATA_DIR = 'data'

if TYPE_CHECKING:
  from huggingface_hub import HfApi


def deploy_project(
  hf_space: str,
  project_dir: Optional[str] = None,
  datasets: Optional[list[str]] = None,
  make_datasets_public: Optional[bool] = False,
  concepts: Optional[list[str]] = None,
  skip_data_upload: Optional[bool] = False,
  skip_concept_upload: Optional[bool] = False,
  create_space: Optional[bool] = False,
  load_on_space: Optional[bool] = False,
  hf_space_storage: Optional[Union[Literal['small'], Literal['medium'], Literal['large']]] = None,
  hf_token: Optional[str] = None,
) -> str:
  """Deploy a project to huggingface.

  Args:
    hf_space: The huggingface space. Should be formatted like `SPACE_ORG/SPACE_NAME`.
    project_dir: The project directory to use for the demo. Defaults to `env.LILAC_PROJECT_DIR`.
    datasets: The names of datasets to upload. Defaults to all datasets.
    make_datasets_public: When true, and when --load_on_space is False, sets the HuggingFace
      datasets that reflect local datasets to public. Defaults to false.
    concepts: The names of concepts to upload. Defaults to all concepts.
    skip_data_upload: When true, kicks the server without uploading data.
    skip_concept_upload: When true, skips uploading concepts.
    create_space: When True, creates the HuggingFace space if it doesnt exist. The space will be
      created with the storage type defined by --hf_space_storage.
    load_on_space: When True, loads the datasets from your project in the space and does not upload
      data. NOTE: This could be expensive if your project config locally has embeddings as they will
      be recomputed in HuggingFace.
    hf_space_storage: If defined, sets the HuggingFace space persistent storage type. NOTE: This
      only actually sets the space storage type when creating the space. For more details, see
      https://huggingface.co/docs/hub/spaces-storage
    hf_token: The HuggingFace access token to upload so that the space can access private datasets.
  """
  try:
    from huggingface_hub import CommitOperationAdd, CommitOperationDelete, HfApi
  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )

  project_dir = project_dir or get_project_dir()
  if not project_dir:
    raise ValueError(
      '--project_dir or the environment variable `LILAC_PROJECT_DIR` must be defined.'
    )

  hf_api = HfApi(token=hf_token)

  operations: list[Union[CommitOperationDelete, CommitOperationAdd]] = deploy_project_operations(
    hf_api=hf_api,
    project_dir=project_dir,
    hf_space=hf_space,
    datasets=datasets,
    make_datasets_public=make_datasets_public,
    concepts=concepts,
    skip_data_upload=skip_data_upload,
    skip_concept_upload=skip_concept_upload,
    create_space=create_space,
    load_on_space=load_on_space,
    hf_space_storage=hf_space_storage,
  )

  # Atomically commit all the operations so we don't kick the server multiple times.
  hf_api.create_commit(
    repo_id=hf_space,
    repo_type='space',
    operations=operations,
    commit_message='Push to HF space',
  )

  link = f'https://huggingface.co/spaces/{hf_space}'
  log(f'Done! View your space at {link}')
  return link


def deploy_project_operations(
  hf_api: 'HfApi',
  project_dir: str,
  hf_space: str,
  datasets: Optional[list[str]] = None,
  make_datasets_public: Optional[bool] = False,
  concepts: Optional[list[str]] = None,
  skip_data_upload: Optional[bool] = False,
  skip_concept_upload: Optional[bool] = False,
  create_space: Optional[bool] = False,
  load_on_space: Optional[bool] = False,
  hf_space_storage: Optional[Union[Literal['small'], Literal['medium'], Literal['large']]] = None,
) -> list:
  """The commit operations for a project deployment."""
  try:
    from huggingface_hub import CommitOperationAdd, CommitOperationDelete
    from huggingface_hub.utils._errors import RepositoryNotFoundError
  except ImportError as e:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    ) from e

  operations: list[Union[CommitOperationDelete, CommitOperationAdd]] = []

  try:
    repo_info = hf_api.repo_info(hf_space, repo_type='space')
  except RepositoryNotFoundError as e:
    if not create_space:
      raise ValueError(
        'The huggingface space does not exist! '
        'Please pass --create_space if you wish to create it.'
      ) from e
    repo_info = None
  if repo_info is None:
    log(f'Creating huggingface space https://huggingface.co/spaces/{hf_space}')
    log('The space will be created as private. You can change this from the UI.')

    if hf_space_storage:
      hf_api.request_space_storage(repo_id=hf_space, storage=hf_space_storage)

    hf_api.create_repo(
      repo_id=hf_space,
      private=True,
      space_storage=hf_space_storage,
      repo_type='space',
      space_sdk='docker',
    )

    log(f'Created: https://huggingface.co/spaces/{hf_space}')

  repo_runtime = hf_api.get_space_runtime(repo_id=hf_space)

  log('Deploying project:', project_dir)
  log()

  ##
  ##  Copy the root files for the docker image.
  ##
  log('Copying root files...')
  log()
  # Upload the hf_docker directory.
  hf_docker_dir = str(resources.files('lilac').joinpath('hf_docker'))
  for upload_file in os.listdir(hf_docker_dir):
    operations.append(
      CommitOperationAdd(
        path_in_repo=upload_file, path_or_fileobj=str(os.path.join(hf_docker_dir, upload_file))
      )
    )

  ##
  ##  Create the empty wheel directory. If uploading a local wheel, use scripts.deploy_staging.
  ##
  operations.extend(_make_wheel_dir(hf_api, hf_space))

  ##
  ##  Upload datasets.
  ##
  project_config = read_project_config(project_dir)
  # When datasets aren't explicitly defined, read all datasets and upload them.
  if datasets is None:
    datasets = [f'{d.namespace}/{d.name}' for d in project_config.datasets]

  if not skip_data_upload and not load_on_space:
    lilac_hf_datasets = _upload_datasets(
      api=hf_api,
      project_dir=project_dir,
      hf_space=hf_space,
      datasets=datasets,
      make_datasets_public=make_datasets_public,
    )
  else:
    lilac_hf_datasets = []

  ##
  ##  Upload the HuggingFace application file (README.md) with uploaded datasets so they are synced
  ##  to storage when the docker image boots up.
  ##
  if (lilac_hf_datasets and not skip_data_upload) or load_on_space:
    readme = (
      '---\n'
      + to_yaml(
        {
          'title': 'Lilac',
          'emoji': '🌷',
          'colorFrom': 'purple',
          'colorTo': 'purple',
          'sdk': 'docker',
          'app_port': 5432,
          'datasets': [d for d in lilac_hf_datasets],
        }
      )
      + '\n---'
    )
    readme_filename = 'README.md'
    if hf_api.file_exists(hf_space, readme_filename, repo_type='space'):
      operations.append(CommitOperationDelete(path_in_repo=readme_filename))

    operations.append(
      CommitOperationAdd(path_in_repo=readme_filename, path_or_fileobj=readme.encode())
    )
  ##
  ##  Upload the lilac.yml project configuration.
  ##
  if datasets and not skip_data_upload:
    project_config_filename = f'data/{PROJECT_CONFIG_FILENAME}'
    # Filter datasets that aren't explicitly defined.
    project_config.datasets = [
      dataset
      for dataset in project_config.datasets
      if f'{dataset.namespace}/{dataset.name}' in datasets
    ]
    if hf_api.file_exists(hf_space, project_config_filename, repo_type='space'):
      operations.append(CommitOperationDelete(path_in_repo=project_config_filename))
    operations.append(
      CommitOperationAdd(
        path_in_repo=project_config_filename,
        path_or_fileobj=to_yaml(
          project_config.model_dump(exclude_defaults=True, exclude_none=True, exclude_unset=True)
        ).encode(),
      )
    )

  ##
  ##  Upload concepts.
  ##
  uploaded_concepts: list[str] = []
  if not skip_concept_upload:
    concept_operations, uploaded_concepts = _upload_concepts(hf_space, project_dir, concepts)
    operations.extend(concept_operations)
    operations.extend(_upload_concept_cache(hf_space, project_dir, uploaded_concepts))

  if repo_runtime.storage:
    hf_api.add_space_variable(hf_space, 'LILAC_PROJECT_DIR', '/data')
    hf_api.add_space_variable(hf_space, 'HF_HOME', '/data/.huggingface')
    hf_api.add_space_variable(hf_space, 'XDG_CACHE_HOME', '/data/.cache')
    hf_api.add_space_variable(hf_space, 'TRANSFORMERS_CACHE', '/data/.cache')
  else:
    hf_api.add_space_variable(hf_space, 'LILAC_PROJECT_DIR', './data')
    hf_api.delete_space_variable(hf_space, 'HF_HOME')
    hf_api.delete_space_variable(hf_space, 'XDG_CACHE_HOME')
    hf_api.delete_space_variable(hf_space, 'TRANSFORMERS_CACHE')

  if load_on_space:
    hf_api.add_space_variable(hf_space, 'LILAC_LOAD_ON_START_SERVER', 'true')
  else:
    hf_api.delete_space_variable(hf_space, 'LILAC_LOAD_ON_START_SERVER')

  if hf_api.token:
    hf_api.add_space_secret(hf_space, 'HF_ACCESS_TOKEN', hf_api.token)

  return operations


def _make_wheel_dir(api: Any, hf_space: str) -> list:
  """Creates the wheel directory README. This does not upload local wheels.

  For local wheels, use deploy_local.
  """
  try:
    from huggingface_hub import CommitOperationAdd, CommitOperationDelete, HfApi
  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )

  hf_api: HfApi = api

  operations: list[Union[CommitOperationDelete, CommitOperationAdd]] = []

  # Make an empty readme in py_dist_dir.
  os.makedirs(PY_DIST_DIR, exist_ok=True)
  with open(os.path.join(PY_DIST_DIR, 'README.md'), 'w') as f:
    f.write(
      'This directory is used for locally built whl files.\n'
      'We write a README.md to ensure an empty folder is uploaded when there is no whl.'
    )

  readme_contents = (
    'This directory is used for locally built whl files.\n'
    'We write a README.md to ensure an empty folder is uploaded when there is no whl.'
  ).encode()

  # Remove everything that exists in dist.
  remote_readme_filepath = os.path.join(PY_DIST_DIR, 'README.md')
  if hf_api.file_exists(hf_space, remote_readme_filepath, repo_type='space'):
    operations.append(CommitOperationDelete(path_in_repo=f'{PY_DIST_DIR}/'))

  operations.append(
    # The path in the remote doesn't os.path.join as it is specific to Linux.
    CommitOperationAdd(path_in_repo=f'{PY_DIST_DIR}/README.md', path_or_fileobj=readme_contents)
  )

  return operations


def _upload_concept_cache(hf_space: str, project_dir: str, concepts: Optional[list[str]]) -> list:
  """Adds local cache (model pkl files) to the HuggingFace Space commit."""
  try:
    from huggingface_hub import CommitOperationAdd, CommitOperationDelete, list_files_info
  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )

  operations: list[Union[CommitOperationDelete, CommitOperationAdd]] = []

  concept_cache_dir = os.path.join(get_lilac_cache_dir(project_dir), CONCEPTS_DIR)
  cache_files: list[str] = []
  remote_cache_dir = os.path.join(get_lilac_cache_dir(REMOTE_DATA_DIR), CONCEPTS_DIR)

  files_info = list(list_files_info(hf_space, f'{remote_cache_dir}/', repo_type='space'))
  if files_info:
    operations.append(CommitOperationDelete(path_in_repo=f'{remote_cache_dir}/'))

  if not os.path.exists(concept_cache_dir):
    return operations
  for root, _, files in os.walk(concept_cache_dir):
    *_, namespace, name = root.rsplit('/', maxsplit=2)
    if concepts and f'{namespace}/{name}' not in concepts:
      continue
    for file in files:
      local_filepath = os.path.join(concept_cache_dir, namespace, name, file)
      # The path in the remote doesn't os.path.join as it is specific to Linux.
      remote_filepath = f'{remote_cache_dir}/{namespace}/{name}/{file}'
      cache_files.append(local_filepath)
      operations.append(
        CommitOperationAdd(
          path_in_repo=remote_filepath,
          path_or_fileobj=local_filepath,
        )
      )

  log('Uploading concept cache files:', cache_files)
  log()

  return operations


def _upload_concepts(
  hf_space: str, project_dir: str, concepts: Optional[list[str]] = None
) -> tuple[list, list[str]]:
  """Adds local concepts to the HuggingFace Space commit."""
  try:
    from huggingface_hub import CommitOperationAdd, CommitOperationDelete, list_files_info
  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )

  operations: list[Union[CommitOperationDelete, CommitOperationAdd]] = []

  disk_concepts = [
    # Remove lilac concepts as they're checked in, and not in the
    f'{c.namespace}/{c.name}'
    for c in DiskConceptDB(project_dir).list()
    if c.namespace != 'lilac'
  ]

  # When concepts are not defined, upload all disk concepts.
  if concepts is None:
    concepts = disk_concepts

  for c in concepts:
    if c not in disk_concepts:
      raise ValueError(f'Concept "{c}" not found in disk concepts: {disk_concepts}')

  for c in concepts:
    namespace, name = c.split('/')

    concept_dir = get_concept_output_dir(project_dir, namespace, name)
    remote_concept_dir = get_concept_output_dir(REMOTE_DATA_DIR, namespace, name)

    files_info = list(list_files_info(hf_space, remote_concept_dir, repo_type='space'))
    if files_info:
      operations.append(CommitOperationDelete(path_in_repo=f'{remote_concept_dir}/'))

    for upload_file in os.listdir(concept_dir):
      operations.append(
        CommitOperationAdd(
          # The path in the remote doesn't os.path.join as it is specific to Linux.
          path_in_repo=f'{remote_concept_dir}/{upload_file}',
          path_or_fileobj=os.path.join(concept_dir, upload_file),
        )
      )

  log('Uploading concepts: ', concepts)
  log()
  return operations, concepts


def _upload_datasets(
  api: Any,
  project_dir: str,
  hf_space: str,
  datasets: list[str],
  make_datasets_public: Optional[bool] = False,
) -> list[str]:
  """Uploads local datasets to HuggingFace datasets."""
  if not make_datasets_public:
    make_datasets_public = False
  try:
    from huggingface_hub import HfApi

  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )
  hf_api: HfApi = api

  hf_space_org, hf_space_name = hf_space.split('/')

  log('Uploading datasets: ', datasets)

  lilac_hf_datasets: list[str] = []
  # Upload datasets to HuggingFace. We do this after uploading code to avoid clobbering the data
  # directory.
  # NOTE(nsthorat): This currently doesn't write to persistent storage directly.
  for d in datasets:
    namespace, name = d.split('/')
    dataset_repo_id = get_hf_dataset_repo_id(hf_space_org, hf_space_name, namespace, name)

    upload(
      dataset=d,
      project_dir=project_dir,
      url_or_repo=dataset_repo_id,
      public=make_datasets_public,
      hf_token=hf_api.token,
    )

    lilac_hf_datasets.append(dataset_repo_id)
  return lilac_hf_datasets


def deploy_config(
  hf_space: str,
  config: Config,
  create_space: Optional[bool] = False,
  hf_space_storage: Optional[Union[Literal['small'], Literal['medium'], Literal['large']]] = None,
  hf_token: Optional[str] = None,
) -> str:
  """Deploys a Lilac config object to a HuggingFace Space.

  Data will be loaded on the HuggingFace space.

  Args:
    hf_space: The HuggingFace space to deploy to. Should be in the format
      "org_name/space_name".
    config: The lilac config object to deploy.
    create_space: When True, creates the HuggingFace space if it doesnt exist. The space will be
      created with the storage type defined by --hf_space_storage.
    hf_space_storage: If defined, sets the HuggingFace space persistent storage type. NOTE: This
      only actually sets the space storage type when creating the space. For more details, see
      https://huggingface.co/docs/hub/spaces-storage
    hf_token: The HuggingFace access token to use when making datasets private. This can also be set
      via the `HF_ACCESS_TOKEN` environment flag.
  """
  with tempfile.TemporaryDirectory() as tmp_project_dir:
    # Write the project config to the temp directory.
    write_project_config(tmp_project_dir, config)

    return deploy_project(
      hf_space=hf_space,
      project_dir=tmp_project_dir,
      create_space=create_space,
      load_on_space=True,
      hf_space_storage=hf_space_storage,
      hf_token=hf_token,
    )


def run(cmd: str, capture_output: bool = False) -> subprocess.CompletedProcess[str]:
  """Run a command and return the result."""
  return subprocess.run(cmd, shell=True, check=True, capture_output=capture_output, text=True)
